#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pytest
import pyomo.environ as pyo

from idaes.core import FlowsheetBlock
from watertap.core.solvers import get_solver
from idaes.core.util.scaling import (
    calculate_scaling_factors,
)
from watertap.property_models.multicomp_aq_sol_prop_pack import MCASParameterBlock
from watertap.unit_models.electrolyzer import Electrolyzer
from watertap.unit_models.tests.unit_test_harness import UnitTestHarness

__author__ = "Hunter Barber"

solver = get_solver()
zero = 1e-6
relative_tolerance = 1e-3


def build():
    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = MCASParameterBlock(
        solute_list=[
            "NA+",
            "CL-",
            "CL2-v",
            "H2-v",
            "OH-",
        ],
        mw_data={
            "H2O": 0.018015,
            "NA+": 0.022989,
            "CL-": 0.03545,
            "CL2-v": 0.0709,
            "H2-v": 0.002016,
            "OH-": 0.017007,
        },
        charge={"NA+": 1, "CL-": -1, "OH-": -1},
        ignore_neutral_charge=True,
    )
    m.fs.unit = Electrolyzer(
        property_package=m.fs.properties,
    )

    # shortcut refs
    prop = m.fs.properties
    anolyte_blk = m.fs.unit.anolyte
    catholyte_blk = m.fs.unit.catholyte

    # fix property parameters
    m.fs.properties.dens_mass_const = 1200

    # feed specifications
    anolyte_blk.properties_in[0].pressure.fix(101325)
    anolyte_blk.properties_in[0].temperature.fix(273.15 + 90)
    anolyte_blk.properties_in[0].flow_mol_phase_comp["Liq", "H2O"].fix(5.551)
    anolyte_blk.properties_in[0].flow_mol_phase_comp["Liq", "NA+"].fix(0.3422)
    anolyte_blk.properties_in[0].flow_mol_phase_comp["Liq", "CL-"].fix(0.3422)
    anolyte_blk.properties_in[0].flow_mol_phase_comp["Liq", "CL2-v"].fix(0)
    anolyte_blk.properties_in[0].flow_mol_phase_comp["Liq", "H2-v"].fix(0)
    anolyte_blk.properties_in[0].flow_mol_phase_comp["Liq", "OH-"].fix(0)
    catholyte_blk.properties_in[0].pressure.fix(101325)
    catholyte_blk.properties_in[0].temperature.fix(273.15 + 90)
    catholyte_blk.properties_in[0].flow_mol_phase_comp["Liq", "H2O"].fix(5.551)
    catholyte_blk.properties_in[0].flow_mol_phase_comp["Liq", "NA+"].fix(1.288)
    catholyte_blk.properties_in[0].flow_mol_phase_comp["Liq", "CL-"].fix(0)
    catholyte_blk.properties_in[0].flow_mol_phase_comp["Liq", "CL2-v"].fix(0)
    catholyte_blk.properties_in[0].flow_mol_phase_comp["Liq", "H2-v"].fix(0)
    catholyte_blk.properties_in[0].flow_mol_phase_comp["Liq", "OH-"].fix(1.288)

    # touch properties
    anolyte_blk.properties_in[0].flow_vol_phase
    anolyte_blk.properties_in[0].conc_mass_phase_comp
    anolyte_blk.properties_in[0].conc_mol_phase_comp
    catholyte_blk.properties_in[0].flow_vol_phase
    catholyte_blk.properties_in[0].conc_mass_phase_comp
    catholyte_blk.properties_in[0].conc_mol_phase_comp
    anolyte_blk.properties_out[0].flow_vol_phase
    anolyte_blk.properties_out[0].conc_mass_phase_comp
    anolyte_blk.properties_out[0].conc_mol_phase_comp
    catholyte_blk.properties_out[0].flow_vol_phase
    catholyte_blk.properties_out[0].conc_mass_phase_comp
    catholyte_blk.properties_out[0].conc_mol_phase_comp

    # fix electrolysis reaction variables
    # TODO: transfer the following variables to generic importable blocks
    # membrane properties
    m.fs.unit.membrane_ion_transport_number["Liq", "NA+"].fix(1)
    # anode properties, Cl- --> 0.5 Cl2 + e-
    m.fs.unit.anode_electrochem_potential.fix(1.21)
    m.fs.unit.anode_stoich["Liq", "CL-"].fix(-1)
    m.fs.unit.anode_stoich["Liq", "CL2-v"].fix(0.5)
    # cathode properties, H20 + e- --> 0.5 H2 + OH-
    m.fs.unit.cathode_electrochem_potential.fix(-0.99)
    m.fs.unit.cathode_stoich["Liq", "H2O"].fix(-1)
    m.fs.unit.cathode_stoich["Liq", "H2-v"].fix(0.5)
    m.fs.unit.cathode_stoich["Liq", "OH-"].fix(1)

    # fix design and performance variables
    # membrane properties
    m.fs.unit.membrane_current_density.fix(4000)
    # anode properties
    m.fs.unit.anode_current_density.fix(3000)
    m.fs.unit.anode_overpotential.fix(0.1)  # assumed
    # cathode properties
    m.fs.unit.cathode_current_density.fix(3000)
    m.fs.unit.cathode_overpotential.fix(0.1)  # assumed
    # electrolyzer cell design
    m.fs.unit.current.fix(30000)
    # performance variables
    m.fs.unit.efficiency_current.fix(0.9)
    m.fs.unit.efficiency_voltage.fix(0.8)

    # scaling
    prop.set_default_scaling("flow_mol_phase_comp", 1, index=("Liq", "H2O"))
    prop.set_default_scaling("flow_mol_phase_comp", 1, index=("Liq", "NA+"))
    prop.set_default_scaling("flow_mol_phase_comp", 1, index=("Liq", "CL-"))
    prop.set_default_scaling("flow_mol_phase_comp", 1, index=("Liq", "CL2-v"))
    prop.set_default_scaling("flow_mol_phase_comp", 1, index=("Liq", "H2-v"))
    prop.set_default_scaling("flow_mol_phase_comp", 1, index=("Liq", "OH-"))
    calculate_scaling_factors(m)

    return m


class TestElectrolyzer(UnitTestHarness):
    def configure(self):
        m = build()

        # arguments for UnitTestHarness
        self.default_zero = zero
        self.default_relative_tolerance = relative_tolerance

        # assert unit model values from solution
        self.unit_solutions[m.fs.unit.membrane_area] = 7.500
        self.unit_solutions[m.fs.unit.anode_area] = 10.00
        self.unit_solutions[m.fs.unit.cathode_area] = 10.00
        self.unit_solutions[m.fs.unit.voltage_cell] = 2.750
        self.unit_solutions[m.fs.unit.resistance] = 1.167e-5
        self.unit_solutions[m.fs.unit.power] = 82510
        self.unit_solutions[m.fs.unit.voltage_reversible] = 2.200
        self.unit_solutions[m.fs.unit.electron_flow] = 0.2798
        self.unit_solutions[m.fs.unit.efficiency_power] = 0.7200

        # check flow at outlet
        self.unit_solutions[
            m.fs.unit.anolyte.properties_out[0].flow_mol_phase_comp["Liq", "CL2-v"]
        ] = 0.1399
        self.unit_solutions[
            m.fs.unit.catholyte.properties_out[0].flow_mol_phase_comp["Liq", "NA+"]
        ] = 1.568
        self.unit_solutions[
            m.fs.unit.catholyte.properties_out[0].flow_mol_phase_comp["Liq", "OH-"]
        ] = 1.568

        self.conservation_equality = {
            "Check 1": {
                "in": m.fs.unit.anolyte.properties_in[0].flow_vol_phase["Liq"]
                + m.fs.unit.catholyte.properties_in[0].flow_vol_phase["Liq"],
                "out": m.fs.unit.anolyte.properties_out[0].flow_vol_phase["Liq"]
                + m.fs.unit.catholyte.properties_out[0].flow_vol_phase["Liq"],
            },
        }

        return m

    @pytest.mark.unit
    def test_electroneutrality(self):

        m = build()
        solver.solve(m)

        # check charge balance
        m.fs.unit.anolyte.properties_in[0].assert_electroneutrality(
            tee=True,
            tol=1e-5,
            solve=False,
        )
        m.fs.unit.anolyte.properties_out[0].assert_electroneutrality(
            tee=True,
            tol=1e-5,
            solve=False,
            defined_state=False,
        )
        m.fs.unit.catholyte.properties_in[0].assert_electroneutrality(
            tee=True,
            tol=1e-5,
            solve=False,
        )
        m.fs.unit.catholyte.properties_out[0].assert_electroneutrality(
            tee=True,
            tol=1e-5,
            solve=False,
            defined_state=False,
        )



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
This module contains a zero-order representation of a supercritical salt precipitation unit.
"""

import pyomo.environ as pyo
from pyomo.environ import units as pyunits, Var
from idaes.core import declare_process_block_class
from watertap.core import build_sido, ZeroOrderBaseData

# Some more information about this module
__author__ = "Chenyu Wang"


@declare_process_block_class("SaltPrecipitationZO")
class SaltPrecipitationZOData(ZeroOrderBaseData):
    """
    Zero-Order model for a supercritical salt precipitation unit.
    """

    CONFIG = ZeroOrderBaseData.CONFIG()

    def build(self):
        super().build()

        self._tech_type = "supercritical_salt_precipitation"

        build_sido(self)

        self.flow_mass_in = Var(
            self.flowsheet().time,
            units=pyunits.t / pyunits.hour,
            bounds=(0, None),
            doc="Inlet mass flowrate",
        )

        @self.Constraint(
            self.flowsheet().time,
            doc="Constraint for inlet mass flowrate.",
        )
        def cons_flow_mass(b, t):
            return b.flow_mass_in[t] == pyunits.convert(
                sum(
                    b.properties_in[t].flow_mass_comp[j]
                    for j in b.properties_in[t].component_list
                ),
                to_units=pyunits.t / pyunits.hour,
            )

        self._perf_var_dict["Inlet Mass Flowrate"] = self.flow_mass_in

        self.electricity = Var(
            self.flowsheet().time,
            units=pyunits.kW,
            bounds=(0, None),
            doc="Electricity consumption of unit",
        )

        self._perf_var_dict["Electricity Demand"] = self.electricity

        self.energy_electric_flow_mass = Var(
            units=pyunits.kWh / pyunits.t,
            doc="Electricity intensity with respect to inlet flowrate",
        )

        @self.Constraint(
            self.flowsheet().time,
            doc="Constraint for electricity consumption based on inlet flowrate.",
        )
        def electricity_consumption(b, t):
            return b.electricity[t] == pyunits.convert(
                b.energy_electric_flow_mass * b.flow_mass_in[t], to_units=pyunits.kW
            )

        self._fixed_perf_vars.append(self.energy_electric_flow_mass)
        self._perf_var_dict["Electricity Intensity"] = self.energy_electric_flow_mass

    @property
    def default_costing_method(self):
        return self.cost_supercritical_salt_precipitation

    @staticmethod
    def cost_supercritical_salt_precipitation(blk):
        """
        General method for costing supercritical salt precipitation unit.
        """
        t0 = blk.flowsheet().time.first()

        # Get parameter dict from database
        parameter_dict = blk.unit_model.config.database.get_unit_operation_parameters(
            blk.unit_model._tech_type, subtype=blk.unit_model.config.process_subtype
        )

        # Get costing parameter sub-block for this technology
        A, B, C, D = blk.unit_model._get_tech_parameters(
            blk,
            parameter_dict,
            blk.unit_model.config.process_subtype,
            [
                "installation_factor",
                "equipment_cost",
                "base_flowrate",
                "scaling_exponent",
            ],
        )

        sizing_term = pyo.units.convert(
            (blk.unit_model.flow_mass_in[t0] / C),
            to_units=pyo.units.dimensionless,
        )

        # Add cost variable and constraint
        blk.capital_cost = pyo.Var(
            initialize=1,
            units=blk.config.flowsheet_costing_block.base_currency,
            bounds=(0, None),
            doc="Capital cost of unit operation",
        )

        expr = pyo.units.convert(
            A * B * sizing_term**D,
            to_units=blk.config.flowsheet_costing_block.base_currency,
        )

        blk.costing_package.add_cost_factor(
            blk, parameter_dict["capital_cost"]["cost_factor"]
        )

        blk.capital_cost_constraint = pyo.Constraint(
            expr=blk.capital_cost == blk.cost_factor * expr
        )

        # Register flows
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.electricity[t0], "electricity"
        )



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Tests for zero-order Ion exchange model
"""
import pytest


from pyomo.environ import (
    Block,
    check_optimal_termination,
    ConcreteModel,
    Constraint,
    value,
    Var,
    Param,
)
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from watertap.core.solvers import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import initialization_tester
from idaes.core import UnitModelCostingBlock

from watertap.unit_models.zero_order import IonExchangeZO
from watertap.core.wt_database import Database
from watertap.core.zero_order_properties import WaterParameterBlock
from watertap.costing.zero_order_costing import ZeroOrderCosting

solver = get_solver()


class TestIonExchangeZO_w_default_removal:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.params = WaterParameterBlock(solute_list=["tds", "foo"])

        m.fs.unit = IonExchangeZO(property_package=m.fs.params, database=m.db)

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(10000)
        m.fs.unit.inlet.flow_mass_comp[0, "tds"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "foo"].fix(4)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database is model.db

        assert isinstance(model.fs.unit.lift_height, Param)
        assert isinstance(model.fs.unit.eta_pump, Param)
        assert isinstance(model.fs.unit.eta_motor, Param)
        assert isinstance(model.fs.unit.electricity, Var)
        assert isinstance(model.fs.unit.electricity_consumption, Constraint)
        assert isinstance(model.fs.unit.water_recovery_equation, Constraint)
        assert isinstance(model.fs.unit.solute_treated_equation, Constraint)

        assert isinstance(model.fs.unit.NaCl_dose, Var)
        assert isinstance(model.fs.unit.NaCl_flowrate, Var)
        assert isinstance(model.fs.unit.NaCl_constraint, Constraint)

        assert isinstance(model.fs.unit.resin_replacement, Var)
        assert isinstance(model.fs.unit.resin_demand, Var)
        assert isinstance(model.fs.unit.resin_constraint, Constraint)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("ion_exchange")

        model.fs.unit.load_parameters_from_database(use_default_removal=True)
        assert model.fs.unit.recovery_frac_mass_H2O[0].fixed
        assert (
            model.fs.unit.recovery_frac_mass_H2O[0].value
            == data["recovery_frac_mass_H2O"]["value"]
        )

        for (t, j), v in model.fs.unit.removal_frac_mass_comp.items():
            assert v.fixed
            if j == "foo":
                assert v.value == data["default_removal_frac_mass_comp"]["value"]
            else:
                assert v.value == data["removal_frac_mass_comp"][j]["value"]

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert pytest.approx(10.005, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].flow_vol
        )
        assert pytest.approx(0.099950024, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["tds"]
        )
        assert pytest.approx(0.3998, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["foo"]
        )
        assert pytest.approx(2899.6, rel=1e-5) == value(model.fs.unit.electricity[0])
        assert (
            model.fs.unit.properties_in[0].flow_mass_comp["H2O"].value
            == model.fs.unit.properties_treated[0].flow_mass_comp["H2O"].value
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, model):
        for j in model.fs.params.component_list:
            assert 1e-5 >= abs(
                value(
                    model.fs.unit.inlet.flow_mass_comp[0, j]
                    - model.fs.unit.treated.flow_mass_comp[0, j]
                    - model.fs.unit.byproduct.flow_mass_comp[0, j]
                )
            )

    @pytest.mark.component
    def test_report(self, model):

        model.fs.unit.report()


class TestIonExchangeZO_clinoptilolite:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.params = WaterParameterBlock(solute_list=["ammonium_as_nitrogen", "foo"])

        m.fs.unit = IonExchangeZO(
            property_package=m.fs.params,
            database=m.db,
            process_subtype="clinoptilolite",
        )

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(10000)
        m.fs.unit.inlet.flow_mass_comp[0, "ammonium_as_nitrogen"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "foo"].fix(4)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database is model.db

        assert isinstance(model.fs.unit.lift_height, Param)
        assert isinstance(model.fs.unit.eta_pump, Param)
        assert isinstance(model.fs.unit.eta_motor, Param)
        assert isinstance(model.fs.unit.electricity, Var)
        assert isinstance(model.fs.unit.electricity_consumption, Constraint)
        assert isinstance(model.fs.unit.water_recovery_equation, Constraint)
        assert isinstance(model.fs.unit.solute_treated_equation, Constraint)

        assert isinstance(model.fs.unit.NaCl_dose, Var)
        assert isinstance(model.fs.unit.NaCl_flowrate, Var)
        assert isinstance(model.fs.unit.NaCl_constraint, Constraint)

        assert isinstance(model.fs.unit.resin_replacement, Var)
        assert isinstance(model.fs.unit.resin_demand, Var)
        assert isinstance(model.fs.unit.resin_constraint, Constraint)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters(
            "ion_exchange", subtype=model.fs.unit.config.process_subtype
        )

        model.fs.unit.load_parameters_from_database(use_default_removal=True)
        assert model.fs.unit.recovery_frac_mass_H2O[0].fixed
        assert (
            model.fs.unit.recovery_frac_mass_H2O[0].value
            == data["recovery_frac_mass_H2O"]["value"]
        )

        for (t, j), v in model.fs.unit.removal_frac_mass_comp.items():
            assert v.fixed
            if j == "foo":
                assert v.value == data["default_removal_frac_mass_comp"]["value"]
            else:
                assert v.value == data["removal_frac_mass_comp"][j]["value"]

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert pytest.approx(9.734, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].flow_vol
        )
        assert pytest.approx(1.027336e-4, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["ammonium_as_nitrogen"]
        )
        assert pytest.approx(0.41093, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["foo"]
        )
        assert pytest.approx(3686.71085, rel=1e-5) == value(
            model.fs.unit.electricity[0]
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, model):
        for j in model.fs.params.component_list:
            assert 1e-5 >= abs(
                value(
                    model.fs.unit.inlet.flow_mass_comp[0, j]
                    - model.fs.unit.treated.flow_mass_comp[0, j]
                    - model.fs.unit.byproduct.flow_mass_comp[0, j]
                )
            )

    @pytest.mark.component
    def test_report(self, model):

        model.fs.unit.report()


db = Database()
params = db._get_technology("ion_exchange")


class TestIXZOsubtype:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.params = WaterParameterBlock(solute_list=["tds"])

        m.fs.unit = IonExchangeZO(property_package=m.fs.params, database=db)

        return m

    @pytest.mark.parametrize("subtype", [k for k in params.keys()])
    @pytest.mark.component
    def test_load_parameters(self, model, subtype):
        model.fs.unit.config.process_subtype = subtype
        data = db.get_unit_operation_parameters("ion_exchange", subtype=subtype)

        model.fs.unit.load_parameters_from_database(use_default_removal=True)

        for (t, j), v in model.fs.unit.removal_frac_mass_comp.items():
            if j not in data["removal_frac_mass_comp"].keys():
                assert v.value == data["default_removal_frac_mass_comp"]["value"]
            else:
                assert v.value == data["removal_frac_mass_comp"][j]["value"]


@pytest.mark.parametrize("subtype", [k for k in params.keys() if k != "clinoptilolite"])
@pytest.mark.component
def test_costing_wt3(subtype):
    m = ConcreteModel()
    m.db = Database()

    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.params = WaterParameterBlock(
        solute_list=["sulfur", "toc", "tds", "ammonium_as_nitrogen"]
    )

    m.fs.costing = ZeroOrderCosting()

    m.fs.unit1 = IonExchangeZO(
        property_package=m.fs.params, database=m.db, process_subtype=subtype
    )

    m.fs.unit1.inlet.flow_mass_comp[0, "H2O"].fix(10000)
    m.fs.unit1.inlet.flow_mass_comp[0, "sulfur"].fix(1)
    m.fs.unit1.inlet.flow_mass_comp[0, "toc"].fix(2)
    m.fs.unit1.inlet.flow_mass_comp[0, "tds"].fix(3)
    m.fs.unit1.inlet.flow_mass_comp[0, "ammonium_as_nitrogen"].fix(1)

    m.fs.unit1.load_parameters_from_database(use_default_removal=True)
    assert degrees_of_freedom(m.fs.unit1) == 0

    m.fs.unit1.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)

    assert isinstance(m.fs.costing.ion_exchange, Block)
    assert isinstance(m.fs.costing.ion_exchange.capital_a_parameter, Var)
    assert isinstance(m.fs.costing.ion_exchange.capital_b_parameter, Var)
    assert isinstance(m.fs.costing.ion_exchange.capital_c_parameter, Var)
    assert isinstance(m.fs.costing.ion_exchange.capital_d_parameter, Var)

    assert isinstance(m.fs.unit1.costing.capital_cost, Var)
    assert isinstance(m.fs.unit1.costing.capital_cost_constraint, Constraint)

    assert_units_consistent(m.fs)
    assert degrees_of_freedom(m.fs.unit1) == 0

    assert m.fs.unit1.electricity[0] in m.fs.costing._registered_flows["electricity"]

    assert (
        m.fs.unit1.NaCl_flowrate[0] in m.fs.costing._registered_flows["sodium_chloride"]
    )
    assert (
        m.fs.unit1.resin_demand[0]
        in m.fs.costing._registered_flows["ion_exchange_resin"]
    )


@pytest.mark.component
def test_costing_clinoptilolite():
    m = ConcreteModel()
    m.db = Database()

    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.params = WaterParameterBlock(solute_list=["ammonium_as_nitrogen"])

    m.fs.costing = ZeroOrderCosting()

    m.fs.unit1 = IonExchangeZO(
        property_package=m.fs.params, database=m.db, process_subtype="clinoptilolite"
    )

    m.fs.unit1.inlet.flow_mass_comp[0, "H2O"].fix(10000)
    m.fs.unit1.inlet.flow_mass_comp[0, "ammonium_as_nitrogen"].fix(1)

    m.fs.unit1.load_parameters_from_database(use_default_removal=True)
    assert degrees_of_freedom(m.fs.unit1) == 0

    m.fs.unit1.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)

    assert isinstance(m.fs.costing.ion_exchange, Block)
    assert isinstance(m.fs.costing.ion_exchange.unit_capex, Var)
    assert isinstance(m.fs.costing.ion_exchange.unit_opex, Var)
    assert isinstance(m.fs.unit1.costing.capital_cost, Var)
    assert isinstance(m.fs.unit1.costing.capital_cost_constraint, Constraint)
    assert_units_consistent(m.fs)
    assert degrees_of_freedom(m.fs.unit1) == 0

    assert m.fs.unit1.electricity[0] in m.fs.costing._registered_flows["electricity"]


@pytest.mark.unit
def test_clinoptilolite_no_ammonium_in_solute_list_error():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.params = WaterParameterBlock(solute_list=["foo"])

    with pytest.raises(
        KeyError,
        match="ammonium_as_nitrogen should be defined in "
        "solute_list for this subtype.",
    ):
        m.fs.unit = IonExchangeZO(
            property_package=m.fs.params, database=db, process_subtype="clinoptilolite"
        )



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Flowsheet example full Water Resource Recovery Facility 
(WRRF; a.k.a., wastewater treatment plant) with ASM1 and ADM1.

The flowsheet follows the same formulation as benchmark simulation model no.2 (BSM2)
but comprises different specifications for default values than BSM2.

"""
__author__ = "Alejandro Garciadiego, Adam Atia, Marcus Holly, Chenyu Wang, Ben Knueven, Xinhong Liu,"

import pyomo.environ as pyo

from pyomo.network import Arc, SequentialDecomposition
from watertap.unit_models.anaerobic_digester import AD
from watertap.unit_models.thickener import Thickener
from watertap.unit_models.dewatering import DewateringUnit
from watertap.unit_models.cstr import CSTR
from watertap.unit_models.clarifier import Clarifier

from watertap.unit_models.translators.translator_asm1_adm1 import Translator_ASM1_ADM1
from watertap.unit_models.translators.translator_adm1_asm1 import Translator_ADM1_ASM1

import idaes.logger as idaeslog
from watertap.core.solvers import get_solver
import idaes.core.util.scaling as iscale
from watertap.property_models.unit_specific.anaerobic_digestion.adm1_properties import (
    ADM1ParameterBlock,
)
from watertap.property_models.unit_specific.anaerobic_digestion.adm1_reactions import (
    ADM1ReactionParameterBlock,
)
from idaes.models.unit_models.mixer import MomentumMixingType
from idaes.models.unit_models.separator import SplittingType
from watertap.property_models.unit_specific.anaerobic_digestion.adm1_properties_vapor import (
    ADM1_vaporParameterBlock,
)

from idaes.core import FlowsheetBlock, UnitModelCostingBlock, UnitModelBlockData
from idaes.models.unit_models import (
    Feed,
    Mixer,
    Separator,
    PressureChanger,
    Product,
)

from watertap.unit_models.aeration_tank import AerationTank, ElectricityConsumption
from watertap.property_models.unit_specific.activated_sludge.asm1_properties import (
    ASM1ParameterBlock,
)
from watertap.property_models.unit_specific.activated_sludge.asm1_reactions import (
    ASM1ReactionParameterBlock,
)
from watertap.core.util.initialization import assert_degrees_of_freedom
from watertap.costing import WaterTAPCosting
from watertap.costing.unit_models.clarifier import (
    cost_circular_clarifier,
    cost_primary_clarifier,
)
from pyomo.util.check_units import assert_units_consistent


def main(reactor_volume_equalities=False):
    m = build()
    set_operating_conditions(m)

    assert_degrees_of_freedom(m, 0)
    assert_units_consistent(m)
    initialize_system(m)

    assert_degrees_of_freedom(m, 0)

    results = solve(m)

    add_costing(m)
    m.fs.costing.initialize()
    assert_degrees_of_freedom(m, 0)

    results = solve(m)
    pyo.assert_optimal_termination(results)

    print("\n\n=============SIMULATION RESULTS=============\n\n")
    # display_results(m)
    display_costing(m)

    setup_optimization(m, reactor_volume_equalities=reactor_volume_equalities)
    results = solve(m, tee=True)
    pyo.assert_optimal_termination(results)
    print("\n\n=============OPTIMIZATION RESULTS=============\n\n")
    # display_results(m)
    display_costing(m)
    display_performance_metrics(m)

    return m, results


def build():
    m = pyo.ConcreteModel()

    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.props_ASM1 = ASM1ParameterBlock()
    m.fs.props_ADM1 = ADM1ParameterBlock()
    m.fs.props_vap = ADM1_vaporParameterBlock()
    m.fs.ADM1_rxn_props = ADM1ReactionParameterBlock(property_package=m.fs.props_ADM1)
    m.fs.ASM1_rxn_props = ASM1ReactionParameterBlock(property_package=m.fs.props_ASM1)
    # Feed water stream
    m.fs.FeedWater = Feed(property_package=m.fs.props_ASM1)

    # ==========================================================================
    # Activated Sludge Process
    # ==========================================================================
    # Mixer for inlet water and recycled sludge
    m.fs.MX1 = Mixer(
        property_package=m.fs.props_ASM1,
        inlet_list=["feed_water", "recycle"],
        momentum_mixing_type=MomentumMixingType.equality,
    )
    # First reactor (anoxic) - standard CSTR
    m.fs.R1 = CSTR(
        property_package=m.fs.props_ASM1, reaction_package=m.fs.ASM1_rxn_props
    )
    # Second reactor (anoxic) - standard CSTR
    m.fs.R2 = CSTR(
        property_package=m.fs.props_ASM1, reaction_package=m.fs.ASM1_rxn_props
    )
    # Third reactor (aerobic) - CSTR with injection
    m.fs.R3 = AerationTank(
        property_package=m.fs.props_ASM1,
        reaction_package=m.fs.ASM1_rxn_props,
        electricity_consumption=ElectricityConsumption.calculated,
    )
    # Fourth reactor (aerobic) - CSTR with injection
    m.fs.R4 = AerationTank(
        property_package=m.fs.props_ASM1,
        reaction_package=m.fs.ASM1_rxn_props,
        electricity_consumption=ElectricityConsumption.calculated,
    )
    # Fifth reactor (aerobic) - CSTR with injection
    m.fs.R5 = AerationTank(
        property_package=m.fs.props_ASM1,
        reaction_package=m.fs.ASM1_rxn_props,
        electricity_consumption=ElectricityConsumption.calculated,
    )
    m.fs.SP5 = Separator(
        property_package=m.fs.props_ASM1, outlet_list=["underflow", "overflow"]
    )
    # Clarifier
    # TODO: Replace with more detailed model when available
    m.fs.CL1 = Clarifier(
        property_package=m.fs.props_ASM1,
        outlet_list=["underflow", "effluent"],
        split_basis=SplittingType.componentFlow,
    )
    # Sludge purge splitter
    m.fs.SP6 = Separator(
        property_package=m.fs.props_ASM1,
        outlet_list=["recycle", "waste"],
        split_basis=SplittingType.totalFlow,
    )
    # Mixing sludge recycle and R5 underflow
    m.fs.MX6 = Mixer(
        property_package=m.fs.props_ASM1,
        inlet_list=["clarifier", "reactor"],
        momentum_mixing_type=MomentumMixingType.equality,
    )

    # Product Blocks
    m.fs.Treated = Product(property_package=m.fs.props_ASM1)
    m.fs.Sludge = Product(property_package=m.fs.props_ASM1)
    # Recycle pressure changer - use a simple isothermal unit for now
    m.fs.P1 = PressureChanger(property_package=m.fs.props_ASM1)

    # Link units
    m.fs.stream2 = Arc(source=m.fs.MX1.outlet, destination=m.fs.R1.inlet)
    m.fs.stream3 = Arc(source=m.fs.R1.outlet, destination=m.fs.R2.inlet)
    m.fs.stream4 = Arc(source=m.fs.R2.outlet, destination=m.fs.R3.inlet)
    m.fs.stream5 = Arc(source=m.fs.R3.outlet, destination=m.fs.R4.inlet)
    m.fs.stream6 = Arc(source=m.fs.R4.outlet, destination=m.fs.R5.inlet)
    m.fs.stream7 = Arc(source=m.fs.R5.outlet, destination=m.fs.SP5.inlet)
    m.fs.stream8 = Arc(source=m.fs.SP5.overflow, destination=m.fs.CL1.inlet)
    m.fs.stream9 = Arc(source=m.fs.SP5.underflow, destination=m.fs.MX6.reactor)
    m.fs.stream10 = Arc(source=m.fs.CL1.effluent, destination=m.fs.Treated.inlet)
    m.fs.stream11 = Arc(source=m.fs.CL1.underflow, destination=m.fs.SP6.inlet)
    m.fs.stream13 = Arc(source=m.fs.SP6.recycle, destination=m.fs.MX6.clarifier)
    m.fs.stream14 = Arc(source=m.fs.MX6.outlet, destination=m.fs.P1.inlet)
    m.fs.stream15 = Arc(source=m.fs.P1.outlet, destination=m.fs.MX1.recycle)
    pyo.TransformationFactory("network.expand_arcs").apply_to(m)

    # Oxygen concentration in reactors 3 and 4 is governed by mass transfer
    m.fs.R3.KLa = 7.6
    m.fs.R4.KLa = 5.7

    # ======================================================================
    # Anaerobic digester section
    m.fs.asm_adm = Translator_ASM1_ADM1(
        inlet_property_package=m.fs.props_ASM1,
        outlet_property_package=m.fs.props_ADM1,
        reaction_package=m.fs.ADM1_rxn_props,
        has_phase_equilibrium=False,
        outlet_state_defined=True,
    )

    m.fs.RADM = AD(
        liquid_property_package=m.fs.props_ADM1,
        vapor_property_package=m.fs.props_vap,
        reaction_package=m.fs.ADM1_rxn_props,
        has_heat_transfer=True,
        has_pressure_change=False,
    )

    m.fs.adm_asm = Translator_ADM1_ASM1(
        inlet_property_package=m.fs.props_ADM1,
        outlet_property_package=m.fs.props_ASM1,
        reaction_package=m.fs.ADM1_rxn_props,
        has_phase_equilibrium=False,
        outlet_state_defined=True,
    )

    # ====================================================================
    # Primary Clarifier
    m.fs.CL = Clarifier(
        property_package=m.fs.props_ASM1,
        outlet_list=["underflow", "effluent"],
        split_basis=SplittingType.componentFlow,
    )

    # Thickener
    m.fs.TU = Thickener(property_package=m.fs.props_ASM1)
    # Dewaterer
    m.fs.DU = DewateringUnit(property_package=m.fs.props_ASM1)

    m.fs.MX2 = Mixer(
        property_package=m.fs.props_ASM1,
        inlet_list=["feed_water1", "recycle1"],
        momentum_mixing_type=MomentumMixingType.equality,
    )
    m.fs.MX3 = Mixer(
        property_package=m.fs.props_ASM1,
        inlet_list=["feed_water2", "recycle2"],
        momentum_mixing_type=MomentumMixingType.equality,
    )
    m.fs.MX4 = Mixer(
        property_package=m.fs.props_ASM1,
        inlet_list=["thickener", "clarifier"],
        momentum_mixing_type=MomentumMixingType.equality,
    )

    # Make connections related to AD section
    m.fs.stream2adm = Arc(
        source=m.fs.RADM.liquid_outlet, destination=m.fs.adm_asm.inlet
    )
    m.fs.stream6adm = Arc(source=m.fs.SP6.waste, destination=m.fs.TU.inlet)
    m.fs.stream3adm = Arc(source=m.fs.TU.underflow, destination=m.fs.MX4.thickener)
    m.fs.stream7adm = Arc(source=m.fs.TU.overflow, destination=m.fs.MX3.recycle2)
    m.fs.stream9adm = Arc(source=m.fs.CL.underflow, destination=m.fs.MX4.clarifier)
    m.fs.stream4adm = Arc(source=m.fs.adm_asm.outlet, destination=m.fs.DU.inlet)
    m.fs.stream5adm = Arc(source=m.fs.DU.overflow, destination=m.fs.MX2.recycle1)
    m.fs.stream11adm = Arc(source=m.fs.DU.underflow, destination=m.fs.Sludge.inlet)
    m.fs.stream01 = Arc(source=m.fs.FeedWater.outlet, destination=m.fs.MX2.feed_water1)
    m.fs.stream02 = Arc(source=m.fs.MX2.outlet, destination=m.fs.MX3.feed_water2)
    m.fs.stream03 = Arc(source=m.fs.MX3.outlet, destination=m.fs.CL.inlet)
    m.fs.stream04 = Arc(source=m.fs.CL.effluent, destination=m.fs.MX1.feed_water)
    m.fs.stream10adm = Arc(source=m.fs.MX4.outlet, destination=m.fs.asm_adm.inlet)
    m.fs.stream1adm = Arc(source=m.fs.asm_adm.outlet, destination=m.fs.RADM.inlet)

    pyo.TransformationFactory("network.expand_arcs").apply_to(m)

    # keep handy all the mixers
    m.mixers = (m.fs.MX1, m.fs.MX2, m.fs.MX3, m.fs.MX4, m.fs.MX6)

    return m


def set_operating_conditions(m):
    # Feed Water Conditions
    m.fs.FeedWater.flow_vol.fix(20648 * pyo.units.m**3 / pyo.units.day)
    m.fs.FeedWater.temperature.fix(308.15 * pyo.units.K)
    m.fs.FeedWater.pressure.fix(1 * pyo.units.atm)
    m.fs.FeedWater.conc_mass_comp[0, "S_I"].fix(27 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "S_S"].fix(58 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "X_I"].fix(92 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "X_S"].fix(363 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "X_BH"].fix(50 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "X_BA"].fix(0 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "X_P"].fix(0 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "S_O"].fix(0 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "S_NO"].fix(0 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "S_NH"].fix(23 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "S_ND"].fix(5 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "X_ND"].fix(16 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.alkalinity.fix(7 * pyo.units.mol / pyo.units.m**3)

    # Reactor sizing in activated sludge process
    m.fs.R1.volume.fix(1000 * pyo.units.m**3)
    m.fs.R2.volume.fix(1000 * pyo.units.m**3)
    m.fs.R3.volume.fix(1333 * pyo.units.m**3)
    m.fs.R4.volume.fix(1333 * pyo.units.m**3)
    m.fs.R5.volume.fix(1333 * pyo.units.m**3)

    # Injection rates to Reactors 3, 4 and 5 of the activated sludge process
    for j in m.fs.props_ASM1.component_list:
        if j != "S_O":
            # All components except S_O have no injection
            m.fs.R3.injection[:, :, j].fix(0)
            m.fs.R4.injection[:, :, j].fix(0)
            m.fs.R5.injection[:, :, j].fix(0)
    # Then set injections rates for O2
    m.fs.R3.outlet.conc_mass_comp[:, "S_O"].fix(1.72e-3)
    m.fs.R4.outlet.conc_mass_comp[:, "S_O"].fix(2.43e-3)
    m.fs.R5.outlet.conc_mass_comp[:, "S_O"].fix(4.49e-4)

    # Set fraction of outflow from reactor 5 that goes to recycle
    m.fs.SP5.split_fraction[:, "underflow"].fix(0.6)

    # Secondary clarifier
    # TODO: Update once secondary clarifier with more detailed model available
    m.fs.CL1.split_fraction[0, "effluent", "H2O"].fix(0.48956)
    m.fs.CL1.split_fraction[0, "effluent", "S_I"].fix(0.48956)
    m.fs.CL1.split_fraction[0, "effluent", "S_S"].fix(0.48956)
    m.fs.CL1.split_fraction[0, "effluent", "X_I"].fix(0.00187)
    m.fs.CL1.split_fraction[0, "effluent", "X_S"].fix(0.00187)
    m.fs.CL1.split_fraction[0, "effluent", "X_BH"].fix(0.00187)
    m.fs.CL1.split_fraction[0, "effluent", "X_BA"].fix(0.00187)
    m.fs.CL1.split_fraction[0, "effluent", "X_P"].fix(0.00187)
    m.fs.CL1.split_fraction[0, "effluent", "S_O"].fix(0.48956)
    m.fs.CL1.split_fraction[0, "effluent", "S_NO"].fix(0.48956)
    m.fs.CL1.split_fraction[0, "effluent", "S_NH"].fix(0.48956)
    m.fs.CL1.split_fraction[0, "effluent", "S_ND"].fix(0.48956)
    m.fs.CL1.split_fraction[0, "effluent", "X_ND"].fix(0.00187)
    m.fs.CL1.split_fraction[0, "effluent", "S_ALK"].fix(0.48956)

    m.fs.CL1.surface_area.fix(1500 * pyo.units.m**2)

    # Sludge purge separator
    m.fs.SP6.split_fraction[:, "recycle"].fix(0.985)

    # Outlet pressure from recycle pump
    m.fs.P1.outlet.pressure.fix(101325)

    # Primary Clarifier
    # TODO: Update primary clarifier once more detailed model available
    m.fs.CL.split_fraction[0, "effluent", "H2O"].fix(0.993)
    m.fs.CL.split_fraction[0, "effluent", "S_I"].fix(0.993)
    m.fs.CL.split_fraction[0, "effluent", "S_S"].fix(0.993)
    m.fs.CL.split_fraction[0, "effluent", "X_I"].fix(0.5192)
    m.fs.CL.split_fraction[0, "effluent", "X_S"].fix(0.5192)
    m.fs.CL.split_fraction[0, "effluent", "X_BH"].fix(0.5192)
    m.fs.CL.split_fraction[0, "effluent", "X_BA"].fix(0.5192)
    m.fs.CL.split_fraction[0, "effluent", "X_P"].fix(0.5192)
    m.fs.CL.split_fraction[0, "effluent", "S_O"].fix(0.993)
    m.fs.CL.split_fraction[0, "effluent", "S_NO"].fix(0.993)
    m.fs.CL.split_fraction[0, "effluent", "S_NH"].fix(0.993)
    m.fs.CL.split_fraction[0, "effluent", "S_ND"].fix(0.993)
    m.fs.CL.split_fraction[0, "effluent", "X_ND"].fix(0.5192)
    m.fs.CL.split_fraction[0, "effluent", "S_ALK"].fix(0.993)

    # Anaerobic digester
    m.fs.RADM.volume_liquid.fix(3400)
    m.fs.RADM.volume_vapor.fix(300)
    m.fs.RADM.liquid_outlet.temperature.fix(308.15)

    # Dewatering Unit - fix either HRT or volume.
    m.fs.DU.hydraulic_retention_time.fix(1800 * pyo.units.s)

    # Set specific energy consumption averaged for centrifuge
    m.fs.DU.energy_electric_flow_vol_inlet[0] = 0.069 * pyo.units.kWh / pyo.units.m**3

    # Thickener unit
    m.fs.TU.hydraulic_retention_time.fix(86400 * pyo.units.s)
    m.fs.TU.diameter.fix(10 * pyo.units.m)

    # TODO: resolve the danger of redundant constraint related to pressure equality constraints created in mixer, specifically for isobaric conditions. the mixer initializer will turn these constraints back on
    for mx in m.mixers:
        mx.pressure_equality_constraints[0.0, 2].deactivate()

    for var in m.fs.component_data_objects(pyo.Var, descend_into=True):
        if "flow_vol" in var.name:
            iscale.set_scaling_factor(var, 1e1)
        if "temperature" in var.name:
            iscale.set_scaling_factor(var, 1e-1)
        if "pressure" in var.name:
            iscale.set_scaling_factor(var, 1e-6)
        if "conc_mass_comp" in var.name:
            iscale.set_scaling_factor(var, 1e1)

    iscale.calculate_scaling_factors(m)


def initialize_system(m):
    # Initialize flowsheet
    # Apply sequential decomposition - 1 iteration should suffice
    seq = SequentialDecomposition()
    seq.options.tear_method = "Direct"
    seq.options.iterLim = 1
    seq.options.tear_set = [m.fs.stream2, m.fs.stream10adm]

    G = seq.create_graph(m)
    # The code below shows tear set and initialization order
    order = seq.calculation_order(G)
    print("Initialization Order")
    for o in order:
        print(o[0].name)

    # Initial guesses for flow into first reactor
    tear_guesses1 = {
        "flow_vol": {0: 103531 / 24 / 3600},
        "conc_mass_comp": {
            (0, "S_I"): 0.028,
            (0, "S_S"): 0.012,
            (0, "X_I"): 1.532,
            (0, "X_S"): 0.069,
            (0, "X_BH"): 2.233,
            (0, "X_BA"): 0.167,
            (0, "X_P"): 0.964,
            (0, "S_O"): 0.0011,
            (0, "S_NO"): 0.0073,
            (0, "S_NH"): 0.0072,
            (0, "S_ND"): 0.0016,
            (0, "X_ND"): 0.0040,
        },
        "alkalinity": {0: 0.0052},
        "temperature": {0: 308.15},
        "pressure": {0: 101325},
    }

    tear_guesses2 = {
        "flow_vol": {0: 178 / 24 / 3600},
        "conc_mass_comp": {
            (0, "S_I"): 0.028,
            (0, "S_S"): 0.048,
            (0, "X_I"): 10.362,
            (0, "X_S"): 20.375,
            (0, "X_BH"): 10.210,
            (0, "X_BA"): 0.553,
            (0, "X_P"): 3.204,
            (0, "S_O"): 0.00025,
            (0, "S_NO"): 0.00169,
            (0, "S_NH"): 0.0289,
            (0, "S_ND"): 0.00468,
            (0, "X_ND"): 0.906,
        },
        "alkalinity": {0: 0.00715},
        "temperature": {0: 308.15},
        "pressure": {0: 101325},
    }

    # Pass the tear_guess to the SD tool
    seq.set_guesses_for(m.fs.R1.inlet, tear_guesses1)
    seq.set_guesses_for(m.fs.asm_adm.inlet, tear_guesses2)

    def function(unit):
        unit.initialize(outlvl=idaeslog.INFO_HIGH)

    seq.run(m, function)

    # TODO: resolve the danger of redundant constraint related to pressure equality constraints created in mixer, specifically for isobaric conditions. the mixer initializer will turn these constraints back on
    for mx in m.mixers:
        mx.pressure_equality_constraints[0.0, 2].deactivate()


def add_costing(m):
    m.fs.costing = WaterTAPCosting()
    m.fs.costing.base_currency = pyo.units.USD_2020

    # Costing Blocks
    m.fs.R1.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
    m.fs.R2.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
    m.fs.R3.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
    m.fs.R4.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
    m.fs.R5.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
    m.fs.CL.costing = UnitModelCostingBlock(
        flowsheet_costing_block=m.fs.costing,
        costing_method=cost_primary_clarifier,
    )

    m.fs.CL1.costing = UnitModelCostingBlock(
        flowsheet_costing_block=m.fs.costing,
        costing_method=cost_circular_clarifier,
    )

    m.fs.RADM.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
    m.fs.DU.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
    m.fs.TU.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)

    # TODO: Leaving out mixer costs; consider including later

    # process costing and add system level metrics
    m.fs.costing.cost_process()
    m.fs.costing.add_annual_water_production(m.fs.Treated.properties[0].flow_vol)
    m.fs.costing.add_LCOW(m.fs.FeedWater.properties[0].flow_vol)
    m.fs.costing.add_specific_energy_consumption(m.fs.FeedWater.properties[0].flow_vol)

    m.fs.objective = pyo.Objective(expr=m.fs.costing.LCOW)
    iscale.set_scaling_factor(m.fs.costing.LCOW, 1e3)
    iscale.set_scaling_factor(m.fs.costing.total_capital_cost, 1e-5)

    for block in m.fs.component_objects(pyo.Block, descend_into=True):
        if isinstance(block, UnitModelBlockData) and hasattr(block, "costing"):
            iscale.set_scaling_factor(block.costing.capital_cost, 1e-6)

    iscale.constraint_scaling_transform(m.fs.DU.costing.capital_cost_constraint, 1e-6)
    iscale.constraint_scaling_transform(m.fs.RADM.costing.capital_cost_constraint, 1e-6)


def setup_optimization(m, reactor_volume_equalities=False):

    for i in ["R1", "R2", "R3", "R4", "R5"]:
        reactor = getattr(m.fs, i)
        reactor.volume.unfix()
        reactor.volume.setlb(1)
        # reactor.volume.setub(2000)
    if reactor_volume_equalities:
        add_reactor_volume_equalities(m)
    m.fs.R3.outlet.conc_mass_comp[:, "S_O"].unfix()
    m.fs.R3.outlet.conc_mass_comp[:, "S_O"].setub(8e-3)

    m.fs.R4.outlet.conc_mass_comp[:, "S_O"].unfix()
    m.fs.R4.outlet.conc_mass_comp[:, "S_O"].setub(8e-3)

    m.fs.R5.outlet.conc_mass_comp[:, "S_O"].unfix()
    m.fs.R5.outlet.conc_mass_comp[:, "S_O"].setub(8e-3)

    # Unfix fraction of outflow from reactor 5 that goes to recycle
    m.fs.SP5.split_fraction[:, "underflow"].unfix()
    # m.fs.SP5.split_fraction[:, "underflow"].setlb(0.45)
    m.fs.SP6.split_fraction[:, "recycle"].unfix()

    add_effluent_violations(m)


def add_effluent_violations(m):
    # TODO: update "m" to blk; change ref to m.fs.Treated instead of CL1 effluent
    m.fs.TSS_max = pyo.Var(initialize=0.03, units=pyo.units.kg / pyo.units.m**3)
    m.fs.TSS_max.fix()

    @m.fs.Constraint(m.fs.time)
    def eq_TSS_max(self, t):
        return m.fs.CL1.effluent_state[0].TSS <= m.fs.TSS_max

    m.fs.COD_max = pyo.Var(initialize=0.1, units=pyo.units.kg / pyo.units.m**3)
    m.fs.COD_max.fix()

    @m.fs.Constraint(m.fs.time)
    def eq_COD_max(self, t):
        return m.fs.CL1.effluent_state[0].COD <= m.fs.COD_max

    m.fs.totalN_max = pyo.Var(initialize=0.018, units=pyo.units.kg / pyo.units.m**3)
    m.fs.totalN_max.fix()

    @m.fs.Constraint(m.fs.time)
    def eq_totalN_max(self, t):
        return m.fs.CL1.effluent_state[0].Total_N <= m.fs.totalN_max

    m.fs.BOD5_max = pyo.Var(initialize=0.01, units=pyo.units.kg / pyo.units.m**3)
    m.fs.BOD5_max.fix()

    @m.fs.Constraint(m.fs.time)
    def eq_BOD5_max(self, t):
        return m.fs.CL1.effluent_state[0].BOD5["effluent"] <= m.fs.BOD5_max


def add_reactor_volume_equalities(m):
    # TODO: These constraints were applied for initial optimization of AS reactor volumes; otherwise, volumes drive towards lower bound. Revisit
    @m.fs.Constraint(m.fs.time)
    def Vol_1(self, t):
        return m.fs.R1.volume[0] == m.fs.R2.volume[0]

    @m.fs.Constraint(m.fs.time)
    def Vol_2(self, t):
        return m.fs.R3.volume[0] == m.fs.R4.volume[0]

    @m.fs.Constraint(m.fs.time)
    def Vol_3(self, t):
        return m.fs.R5.volume[0] >= m.fs.R4.volume[0] * 0.5


def solve(blk, solver=None, tee=False):
    if solver is None:
        solver = get_solver()
    results = solver.solve(blk, tee=tee)
    pyo.assert_optimal_termination(results)
    return results


def display_results(m):
    m.display()

    unit_list = [
        "FeedWater",
        "MX1",
        "R1",
        "R2",
        "R3",
        "R4",
        "R5",
        "SP5",
        "CL1",
        "SP6",
        "MX6",
        "Treated",
        "Sludge",
        "P1",
        "asm_adm",
        "RADM",
        "adm_asm",
        "CL",
        "TU",
        "DU",
        "MX2",
        "MX3",
        "MX4",
    ]
    for u in unit_list:
        m.fs.component(u).report()


def display_costing(m):
    print("Levelized cost of water: %.2f $/m3" % pyo.value(m.fs.costing.LCOW))

    print(
        "Total operating cost: %.2f $/yr" % pyo.value(m.fs.costing.total_operating_cost)
    )
    print("Total capital cost: %.2f $" % pyo.value(m.fs.costing.total_capital_cost))

    print(
        "Total annualized cost: %.2f $/yr"
        % pyo.value(m.fs.costing.total_annualized_cost)
    )
    print(
        "capital cost R1",
        pyo.value(m.fs.R1.costing.capital_cost),
        pyo.units.get_units(m.fs.R1.costing.capital_cost),
    )
    print(
        "capital cost R2",
        pyo.value(m.fs.R2.costing.capital_cost),
        pyo.units.get_units(m.fs.R2.costing.capital_cost),
    )
    print(
        "capital cost R3",
        pyo.value(m.fs.R3.costing.capital_cost),
        pyo.units.get_units(m.fs.R3.costing.capital_cost),
    )
    print(
        "capital cost R4",
        pyo.value(m.fs.R4.costing.capital_cost),
        pyo.units.get_units(m.fs.R4.costing.capital_cost),
    )
    print(
        "capital cost R5",
        pyo.value(m.fs.R5.costing.capital_cost),
        pyo.units.get_units(m.fs.R5.costing.capital_cost),
    )
    print(
        "capital cost primary clarifier",
        pyo.value(m.fs.CL.costing.capital_cost),
        pyo.units.get_units(m.fs.CL.costing.capital_cost),
    )
    print(
        "capital cost secondary clarifier",
        pyo.value(m.fs.CL1.costing.capital_cost),
        pyo.units.get_units(m.fs.CL1.costing.capital_cost),
    )
    print(
        "capital cost AD",
        pyo.value(m.fs.RADM.costing.capital_cost),
        pyo.units.get_units(m.fs.RADM.costing.capital_cost),
    )
    print(
        "capital cost dewatering Unit",
        pyo.value(m.fs.DU.costing.capital_cost),
        pyo.units.get_units(m.fs.DU.costing.capital_cost),
    )
    print(
        "capital cost thickener unit",
        pyo.value(m.fs.TU.costing.capital_cost),
        pyo.units.get_units(m.fs.TU.costing.capital_cost),
    )


def display_performance_metrics(m):
    print(
        "Specific energy consumption with respect to influent flowrate: %.1f kWh/m3"
        % pyo.value(m.fs.costing.specific_energy_consumption)
    )

    print(
        "electricity consumption R3",
        pyo.value(m.fs.R3.electricity_consumption[0]),
        pyo.units.get_units(m.fs.R3.electricity_consumption[0]),
    )
    print(
        "electricity consumption R4",
        pyo.value(m.fs.R4.electricity_consumption[0]),
        pyo.units.get_units(m.fs.R4.electricity_consumption[0]),
    )
    print(
        "electricity consumption R5",
        pyo.value(m.fs.R5.electricity_consumption[0]),
        pyo.units.get_units(m.fs.R5.electricity_consumption[0]),
    )
    print(
        "electricity consumption primary clarifier",
        pyo.value(m.fs.CL.electricity_consumption[0]),
        pyo.units.get_units(m.fs.CL.electricity_consumption[0]),
    )
    print(
        "electricity consumption secondary clarifier",
        pyo.value(m.fs.CL1.electricity_consumption[0]),
        pyo.units.get_units(m.fs.CL1.electricity_consumption[0]),
    )
    print(
        "electricity consumption AD",
        pyo.value(m.fs.RADM.electricity_consumption[0]),
        pyo.units.get_units(m.fs.RADM.electricity_consumption[0]),
    )
    print(
        "electricity consumption dewatering Unit",
        pyo.value(m.fs.DU.electricity_consumption[0]),
        pyo.units.get_units(m.fs.DU.electricity_consumption[0]),
    )
    print(
        "electricity consumption thickening Unit",
        pyo.value(m.fs.TU.electricity_consumption[0]),
        pyo.units.get_units(m.fs.TU.electricity_consumption[0]),
    )
    print(
        "Influent flow",
        pyo.value(m.fs.FeedWater.flow_vol[0]),
        pyo.units.get_units(m.fs.FeedWater.flow_vol[0]),
    )
    print(
        "flow into R3",
        pyo.value(m.fs.R3.control_volume.properties_in[0].flow_vol),
        pyo.units.get_units(m.fs.R3.control_volume.properties_in[0].flow_vol),
    )
    print(
        "flow into RADM",
        pyo.value(m.fs.RADM.liquid_phase.properties_in[0].flow_vol),
        pyo.units.get_units(m.fs.RADM.liquid_phase.properties_in[0].flow_vol),
    )


if __name__ == "__main__":
    m, results = main()



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

"""
This file contains methods to convert WaterTAP naming conventions to OLI
and generate molecular weight and charge dictionaries from molecular formulae.

It calculates molecular weights using the periodic_table.csv from:
https://gist.github.com/GoodmanSciences/c2dd862cd38f21b0ad36b8f96b4bf1ee.
"""

__author__ = "Paul Vecchiarelli, Ben Knueven, Adam Atia"

from collections import namedtuple
from re import findall

from watertap.core.util.chemistry import (
    get_charge,
    get_charge_group,
    get_molar_mass,
)

# TODO: consider replacing some functionality with molmass: https://pypi.org/project/molmass

OLIName = namedtuple(
    "OLIName", ["oli_name", "watertap_name", "charge", "charge_group", "molar_mass"]
)


def watertap_to_oli(watertap_name: str) -> OLIName:
    """
    This method creates a named tuple
    which can be passed directly into OLI
    or into MCAS property models.

    :param watertap_name: string name of substance in WaterTAP format, i.e., B[OH]4_-

    :return OLIName: named tuple containing attributes derived from molecular formula
    """

    c = findall(r"[A-Z]", watertap_name)
    if len(c) == 0:
        raise IOError(
            f" At least 1 uppercase letter is required to specify a molecule, not '{watertap_name}'."
        )

    oli_name = get_oli_name(watertap_name)
    charge = get_charge(watertap_name)
    charge_group = get_charge_group(charge)
    molar_mass = get_molar_mass(watertap_name)
    return OLIName(oli_name, watertap_name, charge, charge_group, molar_mass)


def get_oli_name(watertap_name: str) -> str:
    """
    Converts an WaterTAP formatted name, i.e., "Na_+"
    into an OLI formatted name, i.e., "NAION".

    :param watertap_name: string name of a solute in WaterTAP format

    :return oli_name: string name of a solute in OLI format
    """

    exclude_items = ["temperature", "pressure", "volume"]
    if watertap_name.lower() in exclude_items:
        return watertap_name
    if hasattr(watertap_name, "oli_name"):
        return watertap_name
    components = watertap_name.split("_")
    if len(components) == 0:
        raise IOError(f" Unable to parse solute '{watertap_name}'.")
    if len(components) == 1:
        molecule = components[0]
    elif len(components) == 2:
        molecule = components[0] + "ION"
    oli_name = molecule.replace("[", "").replace("]", "").upper()
    return oli_name


def get_oli_names(source: dict):
    """
    Updates source dictionary with data to populate MCAS property model.

    :param source: dictionary containing WaterTAP names as keys

    :return source: dictionary with OLIName named tuples as keys
    """

    source = dict(
        map(lambda k, v: (watertap_to_oli(k), v), source.keys(), source.values())
    )
    return source


def oli_reverse_lookup(oli_name: str, names_db) -> OLIName:
    """
    Looks up WaterTAP formatted name for solute in OLI format, if listed in names_db dictionary.

    :param oli_name: string name of a solute in OLI format

    :return watertap_name: string name of a solute in WaterTAP format
    """

    if oli_name in names_db:
        return names_db[oli_name]
    else:
        raise IOError(
            f" Component {oli_name} not found in names_db."
            + " Update this dictionary to hard code additional OLI names."
        )


"""
Here follows a dictionary of OLI names and their WaterTAP counterparts.

It functions to aid reverse lookup, i.e., if a name is already in OLI format,
the necessary data can still be extracted.

TODO: method to add novel (valid) names to names_db
"""

names_db = {
    "NAION": "Na_+",
    "CLION": "Cl_-",
    "SO4ION": "SO4_2-",
    "MGION": "Mg_2+",
    "CAION": "Ca_2+",
    "KION": "K_+",
    "HCO3ION": "HCO3_-",
    "NA2CO3": "Na2CO3",
    "CO2": "CO2",
    "H2O": "H2O",
}



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

# Import Pyomo libraries
from pyomo.environ import (
    Set,
    Var,
    Param,
    Suffix,
    Reals,
    Reference,
    TransformationFactory,
    units as pyunits,
)
from pyomo.network import Arc
from pyomo.common.config import ConfigBlock, ConfigValue, In

# Import IDAES cores
from idaes.core import (
    ControlVolume0DBlock,
    declare_process_block_class,
    MaterialBalanceType,
    EnergyBalanceType,
    MomentumBalanceType,
    UnitModelBlockData,
    useDefault,
    MaterialFlowBasis,
)
from idaes.core.util.initialization import propagate_state
from idaes.models.unit_models import Separator
from idaes.models.unit_models.separator import (
    SplittingType,
    EnergySplittingType,
)
from watertap.core.solvers import get_solver
from idaes.core.util.tables import create_stream_table_dataframe
from idaes.core.util.config import is_physical_parameter_block
import idaes.core.util.scaling as iscale
import idaes.logger as idaeslog
from watertap.costing.unit_models.stoichiometric_reactor import (
    cost_stoichiometric_reactor,
)

__author__ = "Tim Bartholomew, Alexander V. Dudchenko"

_log = idaeslog.getLogger(__name__)


@declare_process_block_class("StoichiometricReactor")
class StoichiometricReactorData(UnitModelBlockData):
    """
    StoichiometricReactor - users must provide equations for the solids formed
    """

    CONFIG = ConfigBlock()

    CONFIG.declare(
        "dynamic",
        ConfigValue(
            domain=In([False]),
            default=False,
            description="Dynamic model flag - must be False",
            doc="""Indicates whether this model will be dynamic or not,
    **default** = False. NF units do not support dynamic
    behavior.""",
        ),
    )
    CONFIG.declare(
        "has_holdup",
        ConfigValue(
            default=False,
            domain=In([False]),
            description="Holdup construction flag - must be False",
            doc="""Indicates whether holdup terms should be constructed or not.
    **default** - False. NF units do not have defined volume, thus
    this must be False.""",
        ),
    )
    CONFIG.declare(
        "material_balance_type",
        ConfigValue(
            default=MaterialBalanceType.useDefault,
            domain=In(MaterialBalanceType),
            description="Material balance construction flag",
            doc="""Indicates what type of mass balance should be constructed,
    **default** - MaterialBalanceType.useDefault.
    **Valid values:** {
    **MaterialBalanceType.useDefault - refer to property package for default
    balance type
    **MaterialBalanceType.none** - exclude material balances,
    **MaterialBalanceType.componentPhase** - use phase component balances,
    **MaterialBalanceType.componentTotal** - use total component balances,
    **MaterialBalanceType.elementTotal** - use total element balances,
    **MaterialBalanceType.total** - use total material balance.}""",
        ),
    )
    CONFIG.declare(
        "energy_balance_type",
        ConfigValue(
            default=EnergyBalanceType.useDefault,
            domain=In(EnergyBalanceType),
            description="Energy balance construction flag",
            doc="""Indicates what type of energy balance should be constructed,
    **default** - EnergyBalanceType.useDefault.
    **Valid values:** {
    **EnergyBalanceType.useDefault - refer to property package for default
    balance type
    **EnergyBalanceType.none** - exclude energy balances,
    **EnergyBalanceType.enthalpyTotal** - single enthalpy balance for material,
    **EnergyBalanceType.enthalpyPhase** - enthalpy balances for each phase,
    **EnergyBalanceType.energyTotal** - single energy balance for material,
    **EnergyBalanceType.energyPhase** - energy balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "momentum_balance_type",
        ConfigValue(
            default=MomentumBalanceType.pressureTotal,
            domain=In(MomentumBalanceType),
            description="Momentum balance construction flag",
            doc="""Indicates what type of momentum balance should be constructed,
    **default** - MomentumBalanceType.pressureTotal.
    **Valid values:** {
    **MomentumBalanceType.none** - exclude momentum balances,
    **MomentumBalanceType.pressureTotal** - single pressure balance for material,
    **MomentumBalanceType.pressurePhase** - pressure balances for each phase,
    **MomentumBalanceType.momentumTotal** - single momentum balance for material,
    **MomentumBalanceType.momentumPhase** - momentum balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "has_pressure_change",
        ConfigValue(
            default=False,
            domain=In([True, False]),
            description="Pressure change term construction flag",
            doc="""Indicates whether terms for pressure change should be
    constructed,
    **default** - False.
    **Valid values:** {
    **True** - include pressure change terms,
    **False** - exclude pressure change terms.}""",
        ),
    )
    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations,
    **default** - useDefault.
    **Valid values:** {
    **useDefault** - use default package from parent model or flowsheet,
    **PhysicalParameterObject** - a PhysicalParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigBlock with arguments to be passed to a property block(s)
    and used when constructing these,
    **default** - None.
    **Valid values:** {
    see property package for documentation.}""",
        ),
    )

    CONFIG.declare(
        "reagent",
        ConfigValue(
            default={},
            domain=dict,
            description="Specification of reagents used in StoichiometricReactor process",
            doc="""A dict of reagents used in the StoichiometricReactor process
            including their molecular weights, and dissolution stoichiometric coefficients for
            the components defined in the property package in the following format:

            .. code-block:: python

                {
                "reagent_name_1":
                {
                "mw": (value, units),
                "density_reagent": (value, units),
                "dissolution_stoichiometric":
                {
                "component_name_1": stoichiometric_coeff,
                "component_name_2": stoichiometric_coeff,
                }
                },
                "reagent_name_2":
                {
                "mw": (value, units),
                "density_reagent": (value, units),
                "dissolution_stoichiometric":
                {
                "component_name_1": stoichiometric_coeff,
                "component_name_2": stoichiometric_coeff,
                }
                },
                }

            """,
        ),
    )

    CONFIG.declare(
        "precipitate",
        ConfigValue(
            default={},
            domain=dict,
            description="Specification of precipitates formed in StoichiometricReactor process",
            doc="""
            A dict of precipitates formed in the StoichiometricReactor process
            including their molecular weights, and precipitation stoichiometric coefficients for
            the components defined in the property package in the following format:

            .. code-block:: python

                {
                "precipitate_name_1":
                {
                "mw": (value, units), 
                "precipitation_stoichiometric":
                {
                "component_name_1": stoichiometric_coeff,
                "component_name_2": stoichiometric_coeff,
                }
                },
                "precipitate_name_2":
                {
                "mw": (value, units),
                "precipitation_stoichiometric":
                {
                "component_name_1": stoichiometric_coeff,
                "component_name_2": stoichiometric_coeff,
                }
                },
                }

            """,
        ),
    )

    def build(self):
        # Call UnitModel.build to setup dynamics
        super().build()

        self.scaling_factor = Suffix(direction=Suffix.EXPORT)

        units_meta = self.config.property_package.get_metadata().get_derived_units
        if len(self.config.reagent.keys()):
            self.reagent_list = Set(initialize=self.config.reagent.keys())
            self.has_dissolution_reaction = True
        else:
            self.has_dissolution_reaction = False
        if len(self.config.precipitate.keys()):
            self.precipitate_list = Set(initialize=self.config.precipitate.keys())
            self.has_precipitation_reaction = True
        else:
            self.has_precipitation_reaction = False
        if (
            self.has_dissolution_reaction == False
            and self.has_precipitation_reaction == False
        ):
            raise TypeError(
                "Stoichiometric chemical reactor requires that at least reagents or precipitants are provided"
            )
        # Add unit parameters
        if self.has_dissolution_reaction:
            self.mw_reagent = Param(
                self.reagent_list,
                units=pyunits.kg / pyunits.mol,
                doc="Molecular weight of reagents",
                mutable=True,
            )
            self.density_reagent = Param(
                self.reagent_list,
                initialize=1000,
                units=units_meta("mass") / units_meta("volume"),
                doc="Density of reagents",
                mutable=True,
            )
            for r in self.reagent_list:
                self.mw_reagent[r] = pyunits.convert(
                    self.config.reagent[r]["mw"],
                    to_units=pyunits.kg / pyunits.mol,
                )
                if self.config.reagent[r].get("density_reagent") is not None:
                    self.density_reagent[r] = pyunits.convert(
                        self.config.reagent[r].get("density_reagent"),
                        to_units=(units_meta("mass") / units_meta("volume")),
                    )
                else:
                    _log.warning(
                        "User did not specify density for reagent {}, using 1kg/L".format(
                            r
                        )
                    )
                    self.density_reagent[r] = pyunits.convert(
                        1 * pyunits.kg / pyunits.liter,
                        to_units=(units_meta("mass") / units_meta("volume")),
                    )
            self.dissolution_stoich_comp = Param(
                self.reagent_list,
                self.config.property_package.component_list,
                initialize=0,
                mutable=True,
                doc="Dissolution stoichiometric coefficients for components in property package",
            )
            for r in self.reagent_list:
                for j in self.config.reagent[r]["dissolution_stoichiometric"].keys():
                    self.dissolution_stoich_comp[r, j] = self.config.reagent[r][
                        "dissolution_stoichiometric"
                    ][j]
            # Add unit variables
            self.reagent_dose = Var(
                self.reagent_list,
                initialize=1,
                domain=Reals,
                units=units_meta("mass") / units_meta("volume"),
                doc="reagent dose",
            )
            self.flow_mass_reagent = Var(
                self.reagent_list,
                initialize=1e-3,
                domain=Reals,
                units=units_meta("mass") / units_meta("time"),
                doc="Mass flowrate of reagent",
            )
            self.flow_mol_reagent = Var(
                self.reagent_list,
                initialize=1e-3,
                domain=Reals,
                units=pyunits.mol / units_meta("time"),
                doc="Mass flowrate of reagent",
            )
            self.flow_vol_reagent = Var(
                self.reagent_list,
                initialize=1e-3,
                domain=Reals,
                units=units_meta("volume") / units_meta("time"),
                doc="Volume flowrate of reagent",
            )
            self.dissolution_reaction_generation_comp = Var(
                self.flowsheet().config.time,
                self.config.property_package.component_list,
                domain=Reals,
                initialize=0,
                units=units_meta("mass") / units_meta("time"),
                doc="Mass of component generated from dissolution",
            )

        if self.has_precipitation_reaction:
            self.waste_mass_frac_precipitate = Var(
                initialize=0.2,
                bounds=(0.0, 1),
                units=pyunits.dimensionless,
                doc="Solid mass fraction in sludge",
            )
            self.mw_precipitate = Param(
                self.precipitate_list,
                units=pyunits.kg / pyunits.mol,
                doc="Molecular weight of precipitate",
                mutable=True,
            )
            for p in self.precipitate_list:
                self.mw_precipitate[p] = pyunits.convert(
                    self.config.precipitate[p]["mw"],
                    to_units=pyunits.kg / pyunits.mol,
                )

            self.precipitation_stoich_comp = Param(
                self.precipitate_list,
                self.config.property_package.component_list,
                initialize=0,
                mutable=True,
                doc="Precipitation stoichiometric coefficients for components in property package",
            )
            for p in self.precipitate_list:
                for j in self.config.precipitate[p][
                    "precipitation_stoichiometric"
                ].keys():
                    self.precipitation_stoich_comp[p, j] = self.config.precipitate[p][
                        "precipitation_stoichiometric"
                    ][j]

            self.flow_mass_precipitate = Var(
                self.precipitate_list,
                initialize=1e-3,
                domain=Reals,
                units=units_meta("mass") / units_meta("time"),
                doc="Mass flowrate of precipitate",
            )
            self.flow_mol_precipitate = Var(
                self.precipitate_list,
                initialize=1e-3,
                domain=Reals,
                units=pyunits.mol / units_meta("time"),
                doc="Mass flowrate of precipitate",
            )
            self.conc_mass_precipitate = Var(
                self.precipitate_list,
                initialize=1,
                domain=Reals,
                units=units_meta("mass") / units_meta("volume"),
                doc="Mass concentration of precipitate",
            )

            self.precipitation_reaction_generation_comp = Var(
                self.flowsheet().config.time,
                self.config.property_package.component_list,
                domain=Reals,
                initialize=0,
                units=units_meta("mass") / units_meta("time"),
                doc="Mass of component generated from precipitation",
            )

        if self.has_dissolution_reaction:
            # Build control volume for dissolution reactor
            self.dissolution_reactor = ControlVolume0DBlock(
                dynamic=False,
                has_holdup=False,
                property_package=self.config.property_package,
                property_package_args=self.config.property_package_args,
            )

            self.dissolution_reactor.add_state_blocks(has_phase_equilibrium=False)

            self.dissolution_reactor.add_material_balances(
                balance_type=self.config.material_balance_type,
                has_mass_transfer=True,
            )

            @self.dissolution_reactor.Constraint(
                self.flowsheet().config.time,
                doc="isothermal energy balance for reactor",
            )
            def eq_isothermal_dissolution(b, t):
                return b.properties_in[t].temperature == b.properties_out[t].temperature

            self.dissolution_reactor.add_momentum_balances(
                balance_type=self.config.momentum_balance_type,
                has_pressure_change=self.config.has_pressure_change,
            )
            if (
                self.config.has_pressure_change is True
                and self.config.momentum_balance_type != "none"
            ):
                self.deltaP = Reference(self.dissolution_reactor.deltaP)
        if self.has_precipitation_reaction:
            self.precipitation_reactor = ControlVolume0DBlock(
                dynamic=False,
                has_holdup=False,
                property_package=self.config.property_package,
                property_package_args=self.config.property_package_args,
            )

            self.precipitation_reactor.add_state_blocks(has_phase_equilibrium=False)

            self.precipitation_reactor.add_material_balances(
                balance_type=self.config.material_balance_type,
                has_mass_transfer=True,
            )

            @self.precipitation_reactor.Constraint(
                self.flowsheet().config.time,
                doc="isothermal energy balance for reactor",
            )
            def eq_isothermal_precipitation(b, t):
                return b.properties_in[t].temperature == b.properties_out[t].temperature

            self.precipitation_reactor.add_momentum_balances(
                balance_type=self.config.momentum_balance_type,
                has_pressure_change=self.config.has_pressure_change,
            )
            # References for control volume
            # pressure change
            if (
                self.config.has_pressure_change is True
                and self.config.momentum_balance_type != "none"
            ):
                self.deltaP = Reference(self.precipitation_reactor.deltaP)

            # Build IDAES separator
            self.separator = Separator(
                property_package=self.config.property_package,
                outlet_list=["treated", "waste"],
                split_basis=SplittingType.phaseFlow,
                energy_split_basis=EnergySplittingType.equal_temperature,
            )
        # Add ports

        if self.has_dissolution_reaction:
            # Add Ports
            self.add_inlet_port(name="inlet", block=self.dissolution_reactor)
            if self.has_precipitation_reaction:
                self.add_port(
                    name="dissolution_reactor_outlet",
                    block=self.dissolution_reactor.properties_out,
                )
                self.add_port(
                    name="precipitation_reactor_inlet",
                    block=self.precipitation_reactor.properties_in,
                )
                self.add_port(
                    name="precipitation_reactor_outlet",
                    block=self.precipitation_reactor.properties_out,
                )
                self.add_port(name="outlet", block=self.separator.treated_state)
                self.add_port(name="waste", block=self.separator.waste_state)
                self.dissolution_to_precipitation_reactor = Arc(
                    source=self.dissolution_reactor_outlet,
                    destination=self.precipitation_reactor_inlet,
                )
                self.precipitation_reactor_separator_arc = Arc(
                    source=self.precipitation_reactor_outlet,
                    destination=self.separator.inlet,
                )
                TransformationFactory("network.expand_arcs").apply_to(self)
            else:
                self.add_port(
                    name="outlet", block=self.dissolution_reactor.properties_out
                )
        elif self.has_precipitation_reaction:
            self.add_inlet_port(
                name="inlet",
                block=self.precipitation_reactor,
            )
            self.add_port(
                name="precipitation_reactor_outlet",
                block=self.precipitation_reactor.properties_out,
            )
            self.add_port(name="outlet", block=self.separator.treated_state)
            self.add_port(name="waste", block=self.separator.waste_state)
            self.precipitation_reactor_separator_arc = Arc(
                source=self.precipitation_reactor_outlet,
                destination=self.separator.inlet,
            )
            TransformationFactory("network.expand_arcs").apply_to(self)

        # Dissolution_reactor performance equations
        if self.has_dissolution_reaction:

            @self.Constraint(
                self.flowsheet().config.time,
                self.config.property_package.phase_list,
                self.config.property_package.component_list,
                doc="Mass balance with reaction terms",
            )
            def eq_dissolution_mass_balance(b, t, p, j):
                if (
                    b.config.property_package.config.material_flow_basis
                    == MaterialFlowBasis.mass
                ):
                    return (
                        b.dissolution_reactor.mass_transfer_term[t, p, j]
                        == b.dissolution_reaction_generation_comp[t, j]
                    )
                elif (
                    b.config.property_package.config.material_flow_basis
                    == MaterialFlowBasis.molar
                ):
                    return (
                        b.dissolution_reactor.mass_transfer_term[t, p, j]
                        * b.config.property_package.mw_comp[j]  #
                        == b.dissolution_reaction_generation_comp[t, j]
                    )

            @self.Constraint(
                self.flowsheet().config.time,
                self.config.property_package.component_list,
                doc="Mass generation from dissolution",
            )
            def eq_dissolution_reaction_generation(b, t, j):
                return b.dissolution_reaction_generation_comp[t, j] == sum(
                    b.flow_mass_reagent[r]
                    * b.dissolution_stoich_comp[r, j]
                    * b.config.property_package.mw_comp[j]
                    / b.mw_reagent[r]
                    for r in self.reagent_list
                )

            @self.Constraint(
                self.flowsheet().config.time,
                self.reagent_list,
                doc="Mole generation from dissolution",
            )
            def eq_flow_mol_reagent(b, t, r):
                return b.flow_mass_reagent[r] / b.mw_reagent[r] == b.flow_mol_reagent[r]

            @self.Constraint(
                self.flowsheet().config.time,
                self.reagent_list,
                doc="Reagent mass flow",
            )
            def eq_flow_mass_reagent(b, t, r):
                return (
                    b.flow_mass_reagent[r]
                    == b.reagent_dose[r]
                    * b.dissolution_reactor.properties_in[t].flow_vol_phase["Liq"]
                )

            @self.Constraint(
                self.flowsheet().config.time,
                self.reagent_list,
                doc="Reagent mass flow",
            )
            def eq_flow_vol_reagent(b, t, r):
                return (
                    b.flow_mass_reagent[r] / b.density_reagent[r]
                    == b.flow_vol_reagent[r]
                )

        if self.has_precipitation_reaction:

            @self.Constraint(
                self.flowsheet().config.time,
                self.config.property_package.phase_list,
                self.config.property_package.component_list,
                doc="Mass balance with reaction terms",
            )
            def eq_precipitation_mass_balance(b, t, p, j):
                if (
                    b.config.property_package.config.material_flow_basis
                    == MaterialFlowBasis.mass
                ):
                    return (
                        b.precipitation_reactor.mass_transfer_term[t, p, j]
                        == b.precipitation_reaction_generation_comp[t, j]
                    )
                elif (
                    b.config.property_package.config.material_flow_basis
                    == MaterialFlowBasis.molar
                ):
                    return (
                        b.precipitation_reactor.mass_transfer_term[t, p, j]
                        * b.config.property_package.mw_comp[j]  #
                        == b.precipitation_reaction_generation_comp[t, j]
                    )

            @self.Constraint(
                self.flowsheet().config.time,
                self.config.property_package.component_list,
                doc="Mass generation from precipitation",
            )
            def eq_precipitation_reaction_generation(b, t, j):
                return b.precipitation_reaction_generation_comp[t, j] == -sum(
                    b.flow_mass_precipitate[p]
                    * b.precipitation_stoich_comp[p, j]
                    * b.config.property_package.mw_comp[j]
                    / b.mw_precipitate[p]
                    for p in self.precipitate_list
                )

            @self.Constraint(
                self.flowsheet().config.time,
                self.precipitate_list,
                doc="Mole generation from precipitation",
            )
            def eq_flow_mol_precipitate(b, t, p):
                return (
                    b.flow_mass_precipitate[p] / b.mw_precipitate[p]
                    == b.flow_mol_precipitate[p]
                )

            @self.Constraint(
                self.flowsheet().config.time,
                self.precipitate_list,
                doc="Precipitate mass flow",
            )
            def eq_flow_mass_precipitate(b, t, p):
                return (
                    b.flow_mass_precipitate[p]
                    == b.conc_mass_precipitate[p]
                    * b.precipitation_reactor.properties_out[t].flow_vol_phase["Liq"]
                )

            @self.Constraint(
                self.flowsheet().config.time,
                doc="Mass generation from precipitation",
            )
            def eq_waste_split(b, t):
                return b.waste_mass_frac_precipitate * (
                    sum(b.flow_mass_precipitate[p] for p in b.precipitate_list)
                    + sum(
                        b.separator.waste_state[t].flow_mass_phase_comp["Liq", j]
                        for j in b.config.property_package.component_list
                    )
                ) == sum(b.flow_mass_precipitate[p] for p in b.precipitate_list)

    def initialize_build(
        self, state_args=None, outlvl=idaeslog.NOTSET, solver=None, optarg=None
    ):
        """
        General wrapper for pressure changer initialization routines

        Keyword Arguments:
            state_args : a dict of arguments to be passed to the property
                         package(s) to provide an initial state for
                         initialization (see documentation of the specific
                         property package) (default = {}).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None)
            solver : str indicating which solver to use during
                     initialization (default = None)

        Returns: None
        """
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")

        opt = get_solver(solver, optarg)

        # ---------------------------------------------------------------------
        # Initialize dissolution reactor if constructed
        if self.has_dissolution_reaction:
            disolve_flags = self.dissolution_reactor.initialize(
                outlvl=outlvl,
                optarg=optarg,
                solver=solver,
                state_args=state_args,
            )
            init_log.info_high("Dissolution reactor Step 1 Complete.")
        if self.has_precipitation_reaction:
            # if we have dossolution reactor propgate state to precip reactor
            if self.has_dissolution_reaction:
                propagate_state(self.dissolution_to_precipitation_reactor)

            precip_flags = self.precipitation_reactor.initialize(
                outlvl=outlvl,
                optarg=optarg,
                solver=solver,
                state_args=state_args,
            )
            if self.has_dissolution_reaction:
                self.precipitation_reactor.release_state(precip_flags, outlvl)
            init_log.info_high("Precipitation reactor Step 1 Complete.")
            propagate_state(self.precipitation_reactor_separator_arc)

            self.separator.split_fraction[0, "waste", "Liq"].fix(
                0.01
            )  # guess split fraction as 1%

            self.separator.initialize(
                outlvl=outlvl,
                optarg=optarg,
                solver=solver,
            )

            self.separator.split_fraction[0, "waste", "Liq"].unfix()
            init_log.info_high("Initialization Step 2 Complete.")

        # ---------------------------------------------------------------------
        # Solve unit

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)
        init_log.info_high("Initialization Step 3 {}.".format(idaeslog.condition(res)))

        # ---------------------------------------------------------------------
        # Release Inlet state
        if self.has_dissolution_reaction:
            self.dissolution_reactor.release_state(disolve_flags, outlvl)
        else:
            self.precipitation_reactor.release_state(precip_flags, outlvl)

        init_log.info("Initialization Complete: {}".format(idaeslog.condition(res)))

    def _get_performance_contents(self, time_point=0):
        var_dict = {}
        expr_dict = {}

        for r in self.reagent_list:
            var_dict["Reagent dose - " + r] = self.reagent_dose[r]
            var_dict["Reagent flow - " + r] = self.flow_mass_reagent[r]

        for p in self.precipitate_list:
            var_dict["Precipitate flow - " + p] = self.flow_mass_precipitate[p]
            var_dict["Precipitate conc - " + p] = self.conc_mass_precipitate[p]

        return {"vars": var_dict, "exprs": expr_dict}

    def _get_stream_table_contents(self, time_point=0):
        return create_stream_table_dataframe(
            {
                "Inlet": self.inlet,
                "Reactor outlet": self.reactor_outlet,
                "Outlet": self.outlet,
                "Waste": self.waste,
            },
            time_point=time_point,
        )

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        # setting scaling factors for variables
        # these variables should have user input, if not there will be a warning

        if self.has_dissolution_reaction:

            for r in self.reagent_list:
                sf = 1 / self.mw_reagent[r].value
                iscale.set_scaling_factor(self.mw_reagent[r], sf)
                if iscale.get_scaling_factor(self.reagent_dose[r]) is None:
                    if iscale.get_scaling_factor(self.flow_mass_reagent[r]) is not None:
                        # scale reagent_dose based on flow_mass_reagent
                        sf = iscale.get_scaling_factor(
                            self.flow_mass_reagent[r]
                        ) / iscale.get_scaling_factor(
                            self.dissolution_reactor.properties_in[0].flow_vol_phase[
                                "Liq"
                            ]
                        )
                        iscale.set_scaling_factor(self.reagent_dose[r], sf)
                    else:
                        # default scaling for reagent_dose with warning
                        sf = iscale.get_scaling_factor(
                            self.reagent_dose[r], default=1, warning=True
                        )
                        iscale.set_scaling_factor(self.reagent_dose[r], sf)
                if iscale.get_scaling_factor(self.flow_vol_reagent[r]) is None:
                    sf = iscale.get_scaling_factor(
                        self.flow_mass_reagent[r], default=1e4, warning=True
                    )
                    sf = sf / self.density_reagent[r].value
                    iscale.set_scaling_factor(self.flow_vol_reagent[r], sf)
            for (t, j), v in self.dissolution_reaction_generation_comp.items():
                if iscale.get_scaling_factor(v) is None:
                    sf = iscale.get_scaling_factor(
                        self.dissolution_reactor.properties_in[
                            t
                        ].get_material_flow_terms("Liq", j)
                    )
                    iscale.set_scaling_factor(v, sf)

            for (t, p, j), con in self.eq_dissolution_mass_balance.items():
                sf = iscale.get_scaling_factor(
                    self.dissolution_reaction_generation_comp[t, j]
                )
                iscale.constraint_scaling_transform(con, sf)

            for (t, j), con in self.eq_dissolution_reaction_generation.items():
                sf = iscale.get_scaling_factor(
                    self.dissolution_reaction_generation_comp[t, j]
                )
                iscale.constraint_scaling_transform(con, sf)

            for (t, r), con in self.eq_flow_mass_reagent.items():
                sf = iscale.get_scaling_factor(
                    self.flow_mass_reagent[r], default=1, warning=True
                )
                iscale.constraint_scaling_transform(con, sf)

            for (t, r), con in self.eq_flow_vol_reagent.items():
                sf = iscale.get_scaling_factor(
                    self.flow_vol_reagent[r], default=1, warning=True
                )
                iscale.constraint_scaling_transform(con, sf)

            for r in self.reagent_list:
                if iscale.get_scaling_factor(self.flow_mass_reagent[r]) is None:
                    # scale flow_mass_reagent based on reagent_dose
                    sf = iscale.get_scaling_factor(
                        self.reagent_dose[r]
                    ) * iscale.get_scaling_factor(
                        self.dissolution_reactor.properties_in[0].flow_vol_phase["Liq"]
                    )
                    iscale.set_scaling_factor(self.flow_mass_reagent[r], sf)
                if iscale.get_scaling_factor(self.flow_mol_reagent[r]) is None:
                    sf = iscale.get_scaling_factor(self.flow_mass_reagent[r])
                    sf = sf / iscale.get_scaling_factor(self.mw_reagent[r])
                    iscale.set_scaling_factor(self.flow_mol_reagent[r], sf)
            for (t, r), con in self.eq_flow_mol_reagent.items():
                sf = iscale.get_scaling_factor(
                    self.flow_mol_reagent[r], default=1, warning=True
                )
                iscale.constraint_scaling_transform(con, sf)
            iscale.constraint_scaling_transform(
                self.dissolution_reactor.eq_isothermal_dissolution[0], 1e-2
            )

        if self.has_precipitation_reaction:
            for p in self.precipitate_list:
                sf = 1 / self.mw_precipitate[p].value
                iscale.set_scaling_factor(self.mw_precipitate[p], sf)
                if iscale.get_scaling_factor(self.flow_mass_precipitate[p]) is None:
                    if (
                        iscale.get_scaling_factor(self.conc_mass_precipitate[p])
                        is not None
                    ):
                        # scale flow_mass_precipitate based on conc_mass_precipitate
                        sf = iscale.get_scaling_factor(
                            self.conc_mass_precipitate[p]
                        ) * iscale.get_scaling_factor(
                            self.precipitation_reactor.properties_out[0].flow_vol_phase[
                                "Liq"
                            ]
                        )
                        iscale.set_scaling_factor(self.flow_mass_precipitate[p], sf)
                    else:
                        # default scaling for reagent_dose with warning
                        sf = iscale.get_scaling_factor(
                            self.flow_mass_precipitate[p], default=1e3, warning=True
                        )
                        iscale.set_scaling_factor(self.flow_mass_precipitate[p], sf)
                    if iscale.get_scaling_factor(self.flow_mass_precipitate[p]) is None:
                        iscale.set_scaling_factor(self.flow_mass_precipitate[p], sf)
                if iscale.get_scaling_factor(self.flow_mol_precipitate[p]) is None:
                    sf = iscale.get_scaling_factor(self.flow_mass_precipitate[p])
                    sf = sf / iscale.get_scaling_factor(self.mw_precipitate[p])

                    iscale.set_scaling_factor(self.flow_mol_precipitate[p], sf)
                iscale.constraint_scaling_transform(
                    self.precipitation_reactor.eq_isothermal_precipitation[0], 1e-2
                )
            for (t, p), con in self.eq_flow_mol_precipitate.items():
                sf = iscale.get_scaling_factor(
                    self.flow_mol_precipitate[p], default=1, warning=True
                )
                iscale.constraint_scaling_transform(con, sf)
            for p in self.precipitate_list:
                if iscale.get_scaling_factor(self.conc_mass_precipitate[p]) is None:
                    sf = iscale.get_scaling_factor(
                        self.flow_mass_precipitate[p]
                    ) / iscale.get_scaling_factor(
                        self.precipitation_reactor.properties_out[0].flow_vol_phase[
                            "Liq"
                        ]
                    )
                    iscale.set_scaling_factor(self.conc_mass_precipitate[p], sf)

            for (t, j), v in self.precipitation_reaction_generation_comp.items():
                if iscale.get_scaling_factor(v) is None:
                    sf = iscale.get_scaling_factor(
                        self.precipitation_reactor.properties_in[
                            t
                        ].get_material_flow_terms("Liq", j)
                    )
                    iscale.set_scaling_factor(v, sf)
            for (t, p, j), con in self.eq_precipitation_mass_balance.items():
                sf = iscale.get_scaling_factor(
                    self.precipitation_reaction_generation_comp[t, j]
                )
                iscale.constraint_scaling_transform(con, sf)

            for (t, j), con in self.eq_precipitation_reaction_generation.items():
                sf = iscale.get_scaling_factor(
                    self.precipitation_reaction_generation_comp[t, j]
                )
                iscale.constraint_scaling_transform(con, sf)

            for (t, p), con in self.eq_flow_mass_precipitate.items():
                sf = iscale.get_scaling_factor(self.flow_mass_precipitate[p])
                iscale.constraint_scaling_transform(con, sf)

            sf = (
                sum(
                    iscale.get_scaling_factor(self.flow_mass_precipitate[p]) ** -1
                    for p in self.precipitate_list
                )
                ** -1
            )
            iscale.constraint_scaling_transform(self.eq_waste_split[0], sf)

    @property
    def default_costing_method(self):
        return cost_stoichiometric_reactor



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Tests for anaerobic digester example.

Verified against results from:

Rosen, C. and Jeppsson, U., 2006.
Aspects on ADM1 Implementation within the BSM2 Framework.
Department of Industrial Electrical Engineering and Automation, Lund University, Lund, Sweden, pp.1-35.

"""
from pyomo.environ import (
    ConcreteModel,
)

from idaes.core import (
    FlowsheetBlock,
)

from watertap.core.solvers import get_solver

from watertap.unit_models.anaerobic_digester import AD
from watertap.property_models.unit_specific.anaerobic_digestion.adm1_properties import (
    ADM1ParameterBlock,
)
from watertap.property_models.unit_specific.anaerobic_digestion.adm1_properties_vapor import (
    ADM1_vaporParameterBlock,
)
from watertap.property_models.unit_specific.anaerobic_digestion.adm1_reactions import (
    ADM1ReactionParameterBlock,
)

from watertap.unit_models.tests.unit_test_harness import UnitTestHarness
import idaes.core.util.scaling as iscale

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()


# -----------------------------------------------------------------------------


def build():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.props = ADM1ParameterBlock()
    m.fs.props_vap = ADM1_vaporParameterBlock()
    m.fs.rxn_props = ADM1ReactionParameterBlock(property_package=m.fs.props)

    m.fs.unit = AD(
        liquid_property_package=m.fs.props,
        vapor_property_package=m.fs.props_vap,
        reaction_package=m.fs.rxn_props,
        has_heat_transfer=True,
        has_pressure_change=False,
    )

    # Set the operating conditions
    m.fs.unit.inlet.flow_vol.fix(170 / 24 / 3600)
    m.fs.unit.inlet.temperature.fix(308.15)
    m.fs.unit.inlet.pressure.fix(101325)

    m.fs.unit.inlet.conc_mass_comp[0, "S_su"].fix(0.01)
    m.fs.unit.inlet.conc_mass_comp[0, "S_aa"].fix(0.001)
    m.fs.unit.inlet.conc_mass_comp[0, "S_fa"].fix(0.001)
    m.fs.unit.inlet.conc_mass_comp[0, "S_va"].fix(0.001)
    m.fs.unit.inlet.conc_mass_comp[0, "S_bu"].fix(0.001)
    m.fs.unit.inlet.conc_mass_comp[0, "S_pro"].fix(0.001)
    m.fs.unit.inlet.conc_mass_comp[0, "S_ac"].fix(0.001)
    m.fs.unit.inlet.conc_mass_comp[0, "S_h2"].fix(1e-8)
    m.fs.unit.inlet.conc_mass_comp[0, "S_ch4"].fix(1e-5)
    m.fs.unit.inlet.conc_mass_comp[0, "S_IC"].fix(0.48)
    m.fs.unit.inlet.conc_mass_comp[0, "S_IN"].fix(0.14)
    m.fs.unit.inlet.conc_mass_comp[0, "S_I"].fix(0.02)

    m.fs.unit.inlet.conc_mass_comp[0, "X_c"].fix(2)
    m.fs.unit.inlet.conc_mass_comp[0, "X_ch"].fix(5)
    m.fs.unit.inlet.conc_mass_comp[0, "X_pr"].fix(20)
    m.fs.unit.inlet.conc_mass_comp[0, "X_li"].fix(5)
    m.fs.unit.inlet.conc_mass_comp[0, "X_su"].fix(0.0)
    m.fs.unit.inlet.conc_mass_comp[0, "X_aa"].fix(0.010)
    m.fs.unit.inlet.conc_mass_comp[0, "X_fa"].fix(0.010)
    m.fs.unit.inlet.conc_mass_comp[0, "X_c4"].fix(0.010)
    m.fs.unit.inlet.conc_mass_comp[0, "X_pro"].fix(0.010)
    m.fs.unit.inlet.conc_mass_comp[0, "X_ac"].fix(0.010)
    m.fs.unit.inlet.conc_mass_comp[0, "X_h2"].fix(0.010)
    m.fs.unit.inlet.conc_mass_comp[0, "X_I"].fix(25)

    m.fs.unit.inlet.cations[0].fix(0.04)
    m.fs.unit.inlet.anions[0].fix(0.02)

    m.fs.unit.volume_liquid.fix(3400)
    m.fs.unit.volume_vapor.fix(300)
    m.fs.unit.liquid_outlet.temperature.fix(308.15)

    # Set scaling factors for badly scaled variables
    iscale.set_scaling_factor(
        m.fs.unit.liquid_phase.mass_transfer_term[0, "Liq", "S_h2"], 1e7
    )

    iscale.calculate_scaling_factors(m.fs.unit)

    return m


class TestAnaerobicDigester(UnitTestHarness):
    def configure(self):
        m = build()

        self.unit_solutions[m.fs.unit.liquid_outlet.pressure[0]] = 101325
        self.unit_solutions[m.fs.unit.liquid_outlet.temperature[0]] = 308.15
        self.unit_solutions[m.fs.unit.liquid_outlet.conc_mass_comp[0, "S_I"]] = (
            0.3287724
        )
        self.unit_solutions[m.fs.unit.liquid_outlet.conc_mass_comp[0, "S_aa"]] = (
            0.00531408
        )
        self.unit_solutions[m.fs.unit.liquid_outlet.conc_mass_comp[0, "S_ac"]] = (
            0.1977833
        )
        self.unit_solutions[m.fs.unit.liquid_outlet.conc_mass_comp[0, "S_bu"]] = (
            0.0132484
        )
        self.unit_solutions[m.fs.unit.liquid_outlet.conc_mass_comp[0, "S_ch4"]] = (
            0.0549707
        )
        self.unit_solutions[m.fs.unit.liquid_outlet.conc_mass_comp[0, "S_fa"]] = (
            0.0986058
        )
        self.unit_solutions[m.fs.unit.liquid_outlet.conc_mass_comp[0, "S_h2"]] = (
            2.35916e-07
        )
        self.unit_solutions[m.fs.unit.liquid_outlet.conc_mass_comp[0, "S_pro"]] = (
            0.0157813
        )
        self.unit_solutions[m.fs.unit.liquid_outlet.conc_mass_comp[0, "S_su"]] = (
            0.01195333
        )
        self.unit_solutions[m.fs.unit.liquid_outlet.conc_mass_comp[0, "S_va"]] = (
            0.011622969
        )
        self.unit_solutions[m.fs.unit.liquid_outlet.conc_mass_comp[0, "X_I"]] = 25.6217
        self.unit_solutions[m.fs.unit.liquid_outlet.conc_mass_comp[0, "X_aa"]] = (
            1.1793147
        )
        self.unit_solutions[m.fs.unit.liquid_outlet.conc_mass_comp[0, "X_ac"]] = (
            0.760653
        )
        self.unit_solutions[m.fs.unit.liquid_outlet.conc_mass_comp[0, "X_c"]] = 0.308718
        self.unit_solutions[m.fs.unit.liquid_outlet.conc_mass_comp[0, "X_c4"]] = (
            0.431974
        )
        self.unit_solutions[m.fs.unit.liquid_outlet.conc_mass_comp[0, "X_ch"]] = (
            0.027947465
        )
        self.unit_solutions[m.fs.unit.liquid_outlet.conc_mass_comp[0, "X_fa"]] = (
            0.2430681
        )
        self.unit_solutions[m.fs.unit.liquid_outlet.conc_mass_comp[0, "X_h2"]] = (
            0.3170629
        )
        self.unit_solutions[m.fs.unit.liquid_outlet.conc_mass_comp[0, "X_li"]] = (
            0.0294834
        )
        self.unit_solutions[m.fs.unit.liquid_outlet.conc_mass_comp[0, "X_pr"]] = (
            0.102574392
        )
        self.unit_solutions[m.fs.unit.liquid_outlet.conc_mass_comp[0, "X_pro"]] = (
            0.137323
        )
        self.unit_solutions[m.fs.unit.liquid_outlet.conc_mass_comp[0, "X_su"]] = (
            0.420219
        )
        self.unit_solutions[m.fs.unit.liquid_outlet.conc_mass_comp[0, "S_IC"]] = (
            1.8320212
        )
        self.unit_solutions[m.fs.unit.liquid_outlet.conc_mass_comp[0, "S_IN"]] = (
            1.8235307
        )
        self.unit_solutions[m.fs.unit.liquid_outlet.anions[0]] = 0.0200033
        self.unit_solutions[m.fs.unit.liquid_outlet.cations[0]] = 0.0400066
        self.unit_solutions[m.fs.unit.vapor_outlet.pressure[0]] = 106659.5225
        self.unit_solutions[m.fs.unit.vapor_outlet.temperature[0]] = 308.15
        self.unit_solutions[m.fs.unit.vapor_outlet.flow_vol[0]] = 0.03249637
        self.unit_solutions[m.fs.unit.vapor_outlet.conc_mass_comp[0, "S_ch4"]] = (
            1.6216465
        )
        self.unit_solutions[m.fs.unit.vapor_outlet.conc_mass_comp[0, "S_co2"]] = (
            0.169417
        )
        self.unit_solutions[m.fs.unit.KH_co2[0]] = 0.02714666
        self.unit_solutions[m.fs.unit.KH_ch4[0]] = 0.001161902
        self.unit_solutions[m.fs.unit.KH_h2[0]] = 0.0007384652
        self.unit_solutions[m.fs.unit.electricity_consumption[0]] = 23.7291667
        self.unit_solutions[m.fs.unit.hydraulic_retention_time[0]] = 1880470.588

        # Conservation check

        self.conservation_equality = {
            "Check 1": {
                "in": m.fs.unit.inlet.flow_vol[0],
                "out": (
                    m.fs.unit.liquid_outlet.flow_vol[0] * m.fs.props.dens_mass
                    + m.fs.unit.vapor_outlet.flow_vol[0] * m.fs.props_vap.dens_mass
                )
                / m.fs.props.dens_mass,
            },
            "Check 2": {
                "in": (
                    m.fs.unit.inlet.flow_vol[0]
                    * m.fs.props.dens_mass
                    * m.fs.props.cp_mass
                    * (m.fs.unit.inlet.temperature[0] - m.fs.props.temperature_ref)
                )
                - (
                    m.fs.unit.liquid_outlet.flow_vol[0]
                    * m.fs.props.dens_mass
                    * m.fs.props.cp_mass
                    * (
                        m.fs.unit.liquid_outlet.temperature[0]
                        - m.fs.props.temperature_ref
                    )
                ),
                "out": -1 * m.fs.unit.liquid_phase.enthalpy_transfer[0],
            },
        }

        return m



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
from parameter_sweep import PredeterminedFixedSample, parameter_sweep
import watertap.flowsheets.RO_with_energy_recovery.RO_with_energy_recovery as RO
from watertap.flowsheets.RO_with_energy_recovery.RO_with_energy_recovery import (
    ERDtype,
)


def set_up_sensitivity():
    outputs = {}

    m = RO.build(erd_type=ERDtype.pump_as_turbine)
    RO.set_operating_conditions(m)
    RO.initialize_system(m)
    RO.solve(m)
    m.fs.feed.properties[0].flow_mass_phase_comp.unfix()
    m.fs.feed.properties[0].flow_vol_phase["Liq"].fix()
    m.fs.feed.properties[0].conc_mass_phase_comp["Liq", "NaCl"].fix()
    RO.optimize_set_up(m)
    m.fs.eq_minimum_water_flux.deactivate()
    RO.solve(m)
    RO.display_system(m)
    RO.display_design(m)

    # uncomment to create outputs

    outputs["LCOW"] = m.fs.costing.LCOW
    # outputs["RO Water Flux"] = m.fs.RO.flux_mass_phase_comp[0, 1, "Liq", "H2O"]
    # outputs["RO Membrane Area"] = m.fs.RO.area
    # outputs["RO Energy Consumption"] = m.fs.costing.specific_energy_consumption
    # outputs["System Capital Cost"] = m.fs.costing.aggregate_capital_cost
    # outputs["RO Capital Cost"] = m.fs.RO.costing.capital_cost
    # outputs["Pump Capital Cost"] = m.fs.P1.costing.capital_cost
    # outputs["ERD Capital Cost"] = m.fs.ERD.costing.capital_cost
    # outputs["RO Operating Cost"] = m.fs.RO.costing.fixed_operating_cost
    # outputs[
    #     "MLC Operating Cost"
    # ] = m.fs.costing.maintenance_labor_chemical_operating_cost
    # outputs["Feed Flow Rate"] = m.fs.feed.properties[0].flow_vol_phase["Liq"]
    # outputs["Permeate Flow Rate"] = m.fs.product.properties[0].flow_vol_phase["Liq"]
    # outputs["Retentate Flow Rate"] = m.fs.disposal.properties[0].flow_vol_phase["Liq"]
    # outputs["RO Operating Pressure"] = m.fs.RO.inlet.pressure[0]
    # outputs["RO Permeate H2O Mass Flow"] = m.fs.RO.permeate.flow_mass_phase_comp[
    #     0, "Liq", "H2O"
    # ]
    # outputs["RO Permeate Salt Mass Flow"] = m.fs.RO.permeate.flow_mass_phase_comp[
    #     0, "Liq", "NaCl"
    # ]
    # outputs["RO Retentate H2O Mass Flow"] = m.fs.RO.retentate.flow_mass_phase_comp[
    #     0, "Liq", "H2O"
    # ]
    # outputs["RO Retentate Salt Mass Flow"] = m.fs.RO.retentate.flow_mass_phase_comp[
    #     0, "Liq", "NaCl"
    # ]

    return outputs, m


def run_analysis(case_num=1, nx=5, interpolate_nan_outputs=True, output_filename=None):

    if output_filename is None:
        output_filename = "sensitivity_" + str(case_num) + ".csv"

    # when from the command line
    case_num = int(case_num)
    interpolate_nan_outputs = bool(interpolate_nan_outputs)

    outputs, m = set_up_sensitivity()

    # choose parameter sweep from case structure
    sweep_params = {}

    if case_num == 1:
        # Need to unfix mass recovery of water (or simply sweep across it instead of recovery_vol)
        m.fs.RO.recovery_mass_phase_comp.unfix()

        sweep_params["mass_concentration"] = PredeterminedFixedSample(
            m.fs.feed.properties[0].conc_mass_phase_comp["Liq", "NaCl"],
            [0.963, 1.927, 4.816],
        )
        sweep_params["volumetric_recovery"] = PredeterminedFixedSample(
            m.fs.RO.recovery_vol_phase[0, "Liq"], [0.7, 0.8, 0.9]
        )
    else:
        raise ValueError(f"{case_num} is not yet implemented")

    global_results = parameter_sweep(
        m,
        sweep_params,
        outputs,
        csv_results_file_name=output_filename,
        optimize_function=RO.solve,
        interpolate_nan_outputs=interpolate_nan_outputs,
    )

    return global_results, sweep_params, m


if __name__ == "__main__":
    results, sweep_params, m = run_analysis()
    print(results)



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################


from idaes.core import (
    declare_process_block_class,
)
from idaes.models.unit_models.heat_exchanger import HeatExchangerData
from watertap.core.solvers import get_solver
import idaes.logger as idaeslog
from watertap.costing.unit_models.heat_exchanger import (
    cost_heat_exchanger,
)
from enum import Enum, auto
from pyomo.common.config import ConfigValue


_log = idaeslog.getLogger(__name__)


__author__ = "Elmira Shamlou"

"""
This unit model uses is based on the IDAES `feedwater_heater_0D` model.
However, the constraints and properties defined in the IDAES unit model do not align with those in the available WaterTAP property packages.
To address this, alternative constraints have been replaced to ensure full condensation based on the WaterTAP properties. 
In addition, a condenser option and its corresponding initialization routine have been added, allowing the user to switch between the condenser and steam heater models.
Note that additional components like desuperheaters, drain mixers, and coolers are not included. If necessary, these can be modeled separately by adding heat exchangers and mixers to the flowsheet.
To do: Consider incorporating as a modification to IDAES' FeedWaterHeater model directly in the IDAES repo"
"""


class Mode(Enum):
    HEATER = auto()
    CONDENSER = auto()


@declare_process_block_class(
    "SteamHeater0D",
    doc="""Feedwater Heater Condensing Section
The feedwater heater condensing section model is a normal 0D heat exchanger
model with an added constraint to calculate the steam flow such that the outlet
of shell is a saturated liquid.""",
)
class SteamHeater0DData(HeatExchangerData):
    CONFIG = HeatExchangerData.CONFIG()
    CONFIG.declare(
        "mode",
        ConfigValue(
            default=Mode.HEATER,
            domain=Mode,
            description="Mode of operation: heater or condenser",
        ),
    )
    CONFIG.declare(
        "estimate_cooling_water",
        ConfigValue(
            default=False,
            domain=bool,
            description="Estimate cooling water flow rate for condenser mode",
        ),
    )

    def build(self):
        super().build()

        @self.Constraint(
            self.flowsheet().time,
            self.hot_side.config.property_package.component_list,
            doc="Mass balance",
        )
        def outlet_liquid_mass_balance(b, t, j):
            lb = b.hot_side.properties_out[t].flow_mass_phase_comp["Vap", j].lb
            b.hot_side.properties_out[t].flow_mass_phase_comp["Vap", j].fix(lb)
            return (
                b.hot_side.properties_in[t].flow_mass_phase_comp["Vap", j]
                + b.hot_side.properties_in[t].flow_mass_phase_comp["Liq", j]
                == b.hot_side.properties_out[t].flow_mass_phase_comp["Liq", j]
            )

        @self.Constraint(self.flowsheet().time, doc="Saturation pressure constraint")
        def outlet_pressure_sat(b, t):
            return (
                b.hot_side.properties_out[t].pressure
                >= b.hot_side.properties_out[t].pressure_sat
            )

    def initialize_build(self, *args, **kwargs):
        """
        Initialization routine for both heater and condenser modes. For condenser mode with cooling water estimation, the initialization is performed based on a specified design temperature rise on the cold side.
        """
        solver = kwargs.get("solver", None)
        optarg = kwargs.get("optarg", {})
        outlvl = kwargs.get("outlvl", idaeslog.NOTSET)
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")

        if self.config.mode == Mode.HEATER:
            self.hot_side_inlet.fix()
            self.cold_side_inlet.fix()
            self.hot_side_outlet.unfix()
            self.cold_side_outlet.unfix()
            self.area.fix()

            self.outlet_liquid_mass_balance.deactivate()
            self.outlet_pressure_sat.deactivate()

            # regular heat exchanger initialization
            super().initialize_build(*args, **kwargs)

            for j in self.hot_side.config.property_package.component_list:
                self.hot_side.properties_out[0].flow_mass_phase_comp["Vap", j].fix(0)

            self.outlet_liquid_mass_balance.activate()
            self.outlet_pressure_sat.activate()

            for j in self.hot_side.config.property_package.component_list:
                self.hot_side_inlet.flow_mass_phase_comp[0, "Vap", j].unfix()

            opt = get_solver(solver, optarg)

            with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                res = opt.solve(self, tee=slc.tee)
            init_log.info(
                "Initialization Complete (w/ extraction calc): {}".format(
                    idaeslog.condition(res)
                )
            )
        elif (
            self.config.mode == Mode.CONDENSER
            and not self.config.estimate_cooling_water
        ):
            # condenser mode without cooling water estimation
            super().initialize_build(*args, **kwargs)
            opt = get_solver(solver, optarg)

            with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                res = opt.solve(self, tee=slc.tee)
            init_log.info("Initialization Complete {}".format(idaeslog.condition(res)))

        elif self.config.mode == Mode.CONDENSER and self.config.estimate_cooling_water:

            # condenser mode with cooling water estimation
            cold_side_outlet_temperature = self.cold_side_outlet.temperature[0].value
            self.hot_side_inlet.fix()
            self.cold_side_inlet.fix()
            self.cold_side_outlet.unfix()

            super().initialize_build(*args, **kwargs)
            self.area.unfix()
            self.cold_side_outlet.temperature[0].fix(cold_side_outlet_temperature)
            self.cold_side.properties_in[0].mass_frac_phase_comp["Liq", "TDS"]
            self.cold_side.properties_in[0].mass_frac_phase_comp["Liq", "TDS"].fix()

            for j in self.cold_side.config.property_package.component_list:
                self.cold_side.properties_in[0].flow_mass_phase_comp["Liq", j].unfix()

            self.outlet_liquid_mass_balance.activate()
            self.outlet_pressure_sat.activate()

            opt = get_solver(solver, optarg)

            with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                res = opt.solve(self, tee=slc.tee)
            init_log.info(
                "Initialization Complete (w/ cooling water estimation): {}".format(
                    idaeslog.condition(res)
                )
            )

            self.cold_side.properties_in[0].mass_frac_phase_comp["Liq", "TDS"].unfix()
            self.cold_side.properties_in[0].flow_mass_phase_comp["Liq", "TDS"].fix()

            opt = get_solver(solver, optarg)

            with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                res = opt.solve(self, tee=slc.tee)
            init_log.info(
                "Initialization Complete (w/ cooling water estimation): {}".format(
                    idaeslog.condition(res)
                )
            )

    @property
    def default_costing_method(self):
        return cost_heat_exchanger



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pyomo.environ as pyo
from idaes.core.util.exceptions import ConfigurationError
from idaes.core.util.misc import StrEnum
from ..util import (
    register_costing_parameter_block,
    make_capital_cost_var,
)


class HCType(StrEnum):
    chiller = "chiller"
    electric_heater = "electric_heater"


def build_electric_heater_cost_param_block(blk):

    blk.cost = pyo.Var(
        initialize=66 / 1000,
        bounds=(0, None),
        doc="Heater unit cost",
        units=pyo.units.USD_2018 / pyo.units.watt,
    )

    blk.HE = pyo.Var(
        initialize=0.99,
        bounds=(0, 1),
        doc="electric heater heat generation efficiency",
    )


def cost_heater_chiller(
    blk, HC_type=HCType.electric_heater, cost_electricity_flow=True
):
    """
    electric heater costing method

    Args:
        HC_type: HCType Enum indicating heating or cooling type,
            default = HCType.electric_heater

        cost_electricity_flow: bool, if True, the heater's heat duty will be
            converted to kW and costed as an electricity, default = True
    """
    if HC_type == HCType.electric_heater:
        cost_electric_heater(blk, cost_electricity_flow)
    elif HC_type == HCType.chiller:
        cost_chiller(blk, cost_electricity_flow)
    else:
        raise ConfigurationError(
            f"{blk.unit_model.name} received invalid argument for heater_type:"
            f" {HC_type}. Argument must be a member of the HeaterType Enum."
        )


@register_costing_parameter_block(
    build_rule=build_electric_heater_cost_param_block,
    parameter_block_name="electric_heater",
)
def cost_electric_heater(blk, cost_electricity_flow=True):
    """
    electric heater costing method

    `TODO: describe equations`

    Args:
        cost_electricity_flow (bool): if True, the heater's heat duty will
            be converted to kW and costed as an electricity. Defaults to True.
    """
    t0 = blk.flowsheet().time.first()
    make_capital_cost_var(blk)
    blk.costing_package.add_cost_factor(blk, "TIC")
    blk.capital_cost_constraint = pyo.Constraint(
        expr=blk.capital_cost
        == blk.cost_factor
        * pyo.units.convert(
            blk.costing_package.electric_heater.cost
            * pyo.units.convert(
                blk.unit_model.heat_duty[t0] / blk.costing_package.electric_heater.HE,
                pyo.units.W,
            ),
            to_units=blk.costing_package.base_currency,
        )
    )
    if cost_electricity_flow:
        # grab lower bound of heat duty
        lb = blk.unit_model.heat_duty[t0].lb
        # set lower bound to 0 to avoid negative defined flow warning when lb is not >= 0
        blk.unit_model.heat_duty.setlb(0)
        blk.costing_package.cost_flow(
            pyo.units.convert(
                blk.unit_model.heat_duty[t0] / blk.costing_package.electric_heater.HE,
                to_units=pyo.units.kW,
            ),
            "electricity",
        )
        # set lower bound back to its original value that was assigned to lb
        blk.unit_model.heat_duty.setlb(lb)


def build_chiller_cost_param_block(blk):

    blk.cost = pyo.Var(
        initialize=200 / 1000,
        bounds=(0, None),
        doc="chiller unit cost",
        units=pyo.units.USD_2018 / pyo.units.watt,
    )

    blk.COP = pyo.Var(
        initialize=7,
        bounds=(0, None),
        doc="Chiller coefficient of performance",
    )


@register_costing_parameter_block(
    build_rule=build_chiller_cost_param_block,
    parameter_block_name="chiller",
)
def cost_chiller(blk, cost_electricity_flow=True):
    """
    chiller costing method

    TODO: describe equations

    Args:
        cost_electricity_flow (bool): if True, the chiller's heat_duty will
            be converted to kW and costed as an electricity. Defaults to True.
    """
    t0 = blk.flowsheet().time.first()
    blk.effective_heat_duty = pyo.Var(
        blk.flowsheet().time,
        domain=pyo.NonNegativeReals,
        initialize=0,
        units=pyo.units.watt,
        doc="Effective chiller heat duty (positive value) ",
    )
    blk.effective_heat_duty_constraint = pyo.Constraint(
        expr=blk.effective_heat_duty[t0] == -blk.unit_model.heat_duty[t0]
    )

    make_capital_cost_var(blk)
    blk.costing_package.add_cost_factor(blk, "TIC")
    blk.capital_cost_constraint = pyo.Constraint(
        expr=blk.capital_cost
        == blk.cost_factor
        * pyo.units.convert(
            blk.costing_package.chiller.cost
            * pyo.units.convert(
                blk.effective_heat_duty[t0] / blk.costing_package.chiller.COP,
                to_units=pyo.units.W,
            ),
            to_units=blk.costing_package.base_currency,
        )
    )
    if cost_electricity_flow:
        blk.costing_package.cost_flow(
            pyo.units.convert(
                blk.effective_heat_duty[t0] / blk.costing_package.chiller.COP,
                to_units=pyo.units.kW,
            ),
            "electricity",
        )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import math
from pyomo.environ import (
    value,
)
import logging

_logger = logging.getLogger(__name__)
handler = logging.StreamHandler()
formatter = logging.Formatter(
    "scale utils %(asctime)s %(levelname)s: %(message)s", "%H:%M:%S"
)
handler.setFormatter(formatter)
_logger.addHandler(handler)
_logger.setLevel(logging.DEBUG)

__author__ = "Alexander V. Dudchenko"


def calc_scale(value, factor=1):
    if value == 0:
        return 10 ** (-1 * math.log(abs(1), 10))
    else:
        return 10 ** (-1 * math.log(abs(value), 10) / factor)


def set_default_scaling(
    block_with_values,
    properties_block,
    component="flow_mol_phase_comp",
    factor=1,
    min_value=None,
    max_value=None,
    log_floor_scale=True,
    inverse_scale=False,
):
    for index in block_with_values.find_component(component):
        val = value(block_with_values.find_component(component)[index])
        if min_value != None and val < min_value:
            val = min_value
        if max_value != None and val > max_value:
            val = max_value
        if log_floor_scale:
            scale = calc_scale(val, factor=factor)
        elif inverse_scale:
            scale = 1 / abs(val)
        else:
            raise TypeError(
                "select type of scaling (log_floor_scale or inverse_scale must be True)"
            )
        _logger.info("Default scale for {}: {} val is: {}".format(index, scale, val))
        properties_block.set_default_scaling(component, scale, index=index)



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Tests for ADM1 thermo property package.
Authors: Chenyu Wang, Marcus Holly, Xinhong Liu
"""

import pytest

from pyomo.environ import ConcreteModel, Param, Suffix, value, Var
from pyomo.util.check_units import assert_units_consistent

from idaes.core import MaterialBalanceType, EnergyBalanceType, MaterialFlowBasis

from watertap.property_models.unit_specific.anaerobic_digestion.modified_adm1_properties import (
    ModifiedADM1ParameterBlock,
    ModifiedADM1StateBlock,
    ModifiedADM1PropertiesScaler,
)
from watertap.property_models.tests.property_test_harness import PropertyAttributeError
from idaes.core.util.model_statistics import (
    fixed_variables_set,
    activated_constraints_set,
)

from watertap.core.solvers import get_solver


# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()


class TestParamBlock(object):
    @pytest.fixture(scope="class")
    def model(self):
        model = ConcreteModel()
        model.params = ModifiedADM1ParameterBlock()

        return model

    @pytest.mark.unit
    def test_build(self, model):
        assert model.params.state_block_class is ModifiedADM1StateBlock

        assert len(model.params.phase_list) == 1
        for i in model.params.phase_list:
            assert i == "Liq"

        assert len(model.params.component_list) == 32
        for i in model.params.component_list:
            assert i in [
                "H2O",
                "S_su",
                "S_aa",
                "S_fa",
                "S_va",
                "S_bu",
                "S_pro",
                "S_ac",
                "S_h2",
                "S_ch4",
                "S_IC",
                "S_IN",
                "S_IP",
                "S_I",
                "X_ch",
                "X_pr",
                "X_li",
                "X_su",
                "X_aa",
                "X_fa",
                "X_c4",
                "X_pro",
                "X_ac",
                "X_h2",
                "X_I",
                "X_PHA",
                "X_PP",
                "X_PAO",
                "S_K",
                "S_Mg",
                "S_cat",
                "S_an",
            ]

        assert isinstance(model.params.cp_mass, Param)
        assert value(model.params.cp_mass) == 4182

        assert isinstance(model.params.dens_mass, Param)
        assert value(model.params.dens_mass) == 997

        assert isinstance(model.params.pressure_ref, Param)
        assert value(model.params.pressure_ref) == 101325

        assert isinstance(model.params.temperature_ref, Param)
        assert value(model.params.temperature_ref) == 298.15

        assert isinstance(model.params.CODtoVSS_XI, Var)
        assert model.params.CODtoVSS_XI.is_fixed()
        assert value(model.params.CODtoVSS_XI) == 1.5686

        assert not hasattr(model, "params.CODtoVSS_XS")

        assert isinstance(model.params.CODtoVSS_XBM, Var)
        assert model.params.CODtoVSS_XBM.is_fixed()
        assert value(model.params.CODtoVSS_XBM) == 1.3072

        assert isinstance(model.params.CODtoVSS_XPHA, Var)
        assert model.params.CODtoVSS_XPHA.is_fixed()
        assert value(model.params.CODtoVSS_XPHA) == 1.9608

        assert isinstance(model.params.CODtoVSS_XCH, Var)
        assert model.params.CODtoVSS_XCH.is_fixed()
        assert value(model.params.CODtoVSS_XCH) == 1.5686

        assert isinstance(model.params.CODtoVSS_XPR, Var)
        assert model.params.CODtoVSS_XPR.is_fixed()
        assert value(model.params.CODtoVSS_XPR) == 1.5686

        assert isinstance(model.params.CODtoVSS_XLI, Var)
        assert model.params.CODtoVSS_XLI.is_fixed()
        assert value(model.params.CODtoVSS_XLI) == 1.5686

        assert isinstance(model.params.ISS_P, Var)
        assert model.params.ISS_P.is_fixed()
        assert value(model.params.ISS_P) == 3.23

        assert isinstance(model.params.f_ISS_BM, Var)
        assert model.params.f_ISS_BM.is_fixed()
        assert value(model.params.f_ISS_BM) == 0.15


class TestStateBlock(object):
    @pytest.fixture(scope="class")
    def model(self):
        model = ConcreteModel()
        model.params = ModifiedADM1ParameterBlock()

        model.props = model.params.build_state_block([1])

        return model

    @pytest.mark.unit
    def test_build(self, model):
        assert model.props[1].default_scaler is ModifiedADM1PropertiesScaler

        assert isinstance(model.props[1].flow_vol, Var)
        assert value(model.props[1].flow_vol) == 1

        assert isinstance(model.props[1].pressure, Var)
        assert value(model.props[1].pressure) == 101325

        assert isinstance(model.props[1].temperature, Var)
        assert value(model.props[1].temperature) == 298.15

        assert isinstance(model.props[1].anions, Var)
        assert value(model.props[1].anions) == 0.02

        assert isinstance(model.props[1].cations, Var)
        assert value(model.props[1].cations) == 0.04

        assert isinstance(model.props[1].conc_mass_comp, Var)

        assert len(model.props[1].conc_mass_comp) == 29
        Comp_dict = {
            "S_su": 8.7,
            "S_aa": 0.0053,
            "S_fa": 10.7,
            "S_va": 0.016,
            "S_bu": 0.016,
            "S_pro": 0.036,
            "S_ac": 0.043,
            "S_h2": 0.011,
            "S_ch4": 1e-9,
            "S_IC": 0.15 * 12,
            "S_IN": 0.15 * 14,
            "S_IP": 0.1 * 31,
            "S_I": 0.027,
            "X_ch": 0.041,
            "X_pr": 0.042,
            "X_li": 0.057,
            "X_su": 1e-9,
            "X_aa": 0.49,
            "X_fa": 1e-9,
            "X_c4": 1e-9,
            "X_pro": 1e-9,
            "X_ac": 1e-9,
            "X_h2": 0.32,
            "X_I": 13,
            "X_PHA": 7.28,
            "X_PP": 0.11,
            "X_PAO": 0.69,
            "S_K": 0.33,
            "S_Mg": 0.34,
        }
        for i in model.props[1].conc_mass_comp:
            assert i in [
                "S_su",
                "S_aa",
                "S_fa",
                "S_va",
                "S_bu",
                "S_pro",
                "S_ac",
                "S_h2",
                "S_ch4",
                "S_IC",
                "S_IN",
                "S_IP",
                "S_I",
                "X_ch",
                "X_pr",
                "X_li",
                "X_su",
                "X_aa",
                "X_fa",
                "X_c4",
                "X_pro",
                "X_ac",
                "X_h2",
                "X_I",
                "X_PHA",
                "X_PP",
                "X_PAO",
                "S_K",
                "S_Mg",
            ]
            assert value(model.props[1].conc_mass_comp[i]) == Comp_dict[i]

        metadata = model.params.get_metadata().properties

        # check that properties are not built if not demanded
        for v in metadata.list_supported_properties():
            if metadata[v.name].method is not None:
                if model.props[1].is_property_constructed(v.name):
                    raise PropertyAttributeError(
                        "Property {v_name} is an on-demand property, but was found "
                        "on the stateblock without being demanded".format(v_name=v.name)
                    )

        # check that properties are built if demanded
        for v in metadata.list_supported_properties():
            if metadata[v.name].method is not None:
                if not hasattr(model.props[1], v.name):
                    raise PropertyAttributeError(
                        "Property {v_name} is an on-demand property, but was not built "
                        "when demanded".format(v_name=v.name)
                    )

    @pytest.mark.unit
    def test_get_material_flow_terms(self, model):
        for p in model.params.phase_list:
            for j in model.params.component_list:
                assert model.props[1].get_material_flow_terms(p, j) is (
                    model.props[1].material_flow_expression[j]
                )

    @pytest.mark.unit
    def test_get_enthalpy_flow_terms(self, model):
        for p in model.params.phase_list:
            assert model.props[1].get_enthalpy_flow_terms(p) is (
                model.props[1].enthalpy_flow_expression
            )

    @pytest.mark.unit
    def test_get_material_density_terms(self, model):
        for p in model.params.phase_list:
            for j in model.params.component_list:
                assert model.props[1].get_material_density_terms(p, j) is (
                    model.props[1].material_density_expression[j]
                )

    @pytest.mark.unit
    def test_get_energy_density_terms(self, model):
        for p in model.params.phase_list:
            assert model.props[1].get_energy_density_terms(p) is (
                model.props[1].energy_density_expression
            )

    @pytest.mark.unit
    def test_default_material_balance_type(self, model):
        assert (
            model.props[1].default_material_balance_type()
            == MaterialBalanceType.componentPhase
        )

    @pytest.mark.unit
    def test_default_energy_balance_type(self, model):
        assert (
            model.props[1].default_energy_balance_type()
            == EnergyBalanceType.enthalpyTotal
        )

    @pytest.mark.unit
    def test_get_material_flow_basis(self, model):
        assert model.props[1].get_material_flow_basis() == MaterialFlowBasis.mass

    @pytest.mark.unit
    def test_define_state_vars(self, model):
        sv = model.props[1].define_state_vars()

        assert len(sv) == 6
        for i in sv:
            assert i in [
                "flow_vol",
                "pressure",
                "temperature",
                "conc_mass_comp",
                "anions",
                "cations",
            ]

    @pytest.mark.unit
    def test_define_port_members(self, model):
        sv = model.props[1].define_port_members()

        assert len(sv) == 6
        for i in sv:
            assert i in [
                "flow_vol",
                "pressure",
                "temperature",
                "conc_mass_comp",
                "anions",
                "cations",
            ]

    @pytest.mark.unit
    def test_define_display_vars(self, model):
        sv = model.props[1].define_display_vars()

        assert len(sv) == 6
        for i in sv:
            assert i in [
                "Volumetric Flowrate",
                "Molar anions",
                "Molar cations",
                "Mass Concentration",
                "Temperature",
                "Pressure",
            ]

    @pytest.mark.component
    def test_initialize(self, model):
        orig_fixed_vars = fixed_variables_set(model)
        orig_act_consts = activated_constraints_set(model)

        model.props.initialize(hold_state=False)

        fin_fixed_vars = fixed_variables_set(model)
        fin_act_consts = activated_constraints_set(model)

        assert len(fin_act_consts) == len(orig_act_consts)
        assert len(fin_fixed_vars) == len(orig_fixed_vars)

        for c in fin_act_consts:
            assert c in orig_act_consts
        for v in fin_fixed_vars:
            assert v in orig_fixed_vars

    @pytest.mark.unit
    def check_units(self, model):
        assert_units_consistent(model)


class TestModifiedADM1PropertiesScaler:
    @pytest.mark.unit
    def test_variable_scaling_routine(self):
        model = ConcreteModel()
        model.params = ModifiedADM1ParameterBlock()

        model.props = model.params.build_state_block([1], defined_state=False)

        scaler = model.props[1].default_scaler()
        assert isinstance(scaler, ModifiedADM1PropertiesScaler)

        scaler.variable_scaling_routine(model.props[1])

        sfx = model.props[1].scaling_factor
        assert len(sfx) == 3
        assert sfx[model.props[1].flow_vol] == pytest.approx(1e5, rel=1e-8)
        assert sfx[model.props[1].pressure] == pytest.approx(1e-6, rel=1e-8)
        assert sfx[model.props[1].temperature] == pytest.approx(1e-1, rel=1e-8)

    @pytest.mark.unit
    def test_constraint_scaling_routine(self):
        model = ConcreteModel()
        model.params = ModifiedADM1ParameterBlock()

        model.props = model.params.build_state_block([1], defined_state=False)

        scaler = model.props[1].default_scaler()
        assert isinstance(scaler, ModifiedADM1PropertiesScaler)

        scaler.constraint_scaling_routine(model.props[1])

    @pytest.mark.unit
    def test_scale_model(self):
        model = ConcreteModel()
        model.params = ModifiedADM1ParameterBlock()

        model.props = model.params.build_state_block([1], defined_state=False)

        scaler = model.props[1].default_scaler()
        assert isinstance(scaler, ModifiedADM1PropertiesScaler)

        scaler.scale_model(model.props[1])

        assert isinstance(model.props[1].scaling_factor, Suffix)

        sfx = model.props[1].scaling_factor
        assert len(sfx) == 3
        assert sfx[model.props[1].flow_vol] == pytest.approx(1e5, rel=1e-8)
        assert sfx[model.props[1].pressure] == pytest.approx(1e-6, rel=1e-8)
        assert sfx[model.props[1].temperature] == pytest.approx(1e-1, rel=1e-8)
